import asyncio
import json
from unittest import IsolatedAsyncioTestCase

from uuid_utils import uuid4

from .....admin.request_context import AdminRequestContext
from .....anoncreds.default.legacy_indy.registry import LegacyIndyRegistry
from .....anoncreds.issuer import AnonCredsIssuer
from .....anoncreds.revocation import AnonCredsRevocation
from .....cache.base import BaseCache
from .....cache.in_memory import InMemoryCache
from .....connections.models.conn_record import ConnRecord
from .....ledger.base import BaseLedger
from .....tests import mock
from .....utils.testing import create_test_profile
from .....wallet.base import BaseWallet
from .....wallet.did_method import SOV, DIDMethods
from .....wallet.key_type import ED25519, KeyTypes
from ....issue_credential.v1_0.tests import REV_REG_ID
from ..manager import TransactionManager, TransactionManagerError
from ..models.transaction_record import TransactionRecord
from ..transaction_jobs import TransactionJob

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "bc-reg"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:tag1"


class TestTransactionManager(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        sigs = [
            (
                "2iNTeFy44WK9zpsPfcwfu489aHWroYh3v8mme9tPyNKn"
                "crk1tVbWKNU4zFvLAbSBwHWxShQSJrhRgoxwaehCaz2j"
            ),
            (
                "3hPr2WgAixcXQRQfCZKnmpY7SkQyQW4cegX7QZMPv6Fv"
                "sNRFV7yW21VaFC5CA3Aze264dkHjX4iZ1495am8fe1qZ"
            ),
        ]
        self.test_messages_attach = f"""{{
            "endorser": "DJGEjaMunDtFtBVrn1qJMT",
            "identifier": "C3nJhruVc7feyB6ckJwhi2",
            "operation": {{
                "data": {{
                    "attr_names": ["score"],
                    "name": "prefs",
                    "version": "1.0"
                }},
                "type": "101"
            }},
            "protocolVersion": 2,
            "reqId": 1613463373859595201,
            "signatures": {{
                "C3nJhruVc7feyB6ckJwhi2": {sigs[0]}
            }}
        }}"""

        self.test_expires_time = "2021-03-29T05:22:19Z"
        self.test_connection_id = "3fa85f64-5717-4562-b3fc-2c963f66afa6"
        self.test_receivers_connection_id = "3fa85f64-5717-4562-b3fc-2c963f66afa7"
        self.test_author_transaction_id = "3fa85f64-5717-4562-b3fc-2c963f66afa7"
        self.test_endorser_transaction_id = "3fa85f64-5717-4562-b3fc-2c963f66afa8"

        self.test_endorsed_message = f"""{{
            "endorser": "DJGEjaMunDtFtBVrn1qJMT",
            "identifier": "C3nJhruVc7feyB6ckJwhi2",
            "operation": {{
                "data": {{
                    "attr_names": ["score"],
                    "name": "prefs",
                    "version": "1.0"
                }},
                "type": "101"
            }},
            "protocolVersion": 2,
            "reqId": 1613463373859595201,
            "signatures": {{
                "C3nJhruVc7feyB6ckJwhi2": {sigs[0]},
                "DJGEjaMunDtFtBVrn1qJMT": {sigs[1]}
            }}
        }}"""

        self.test_signature = f"""{{
            "endorser": "DJGEjaMunDtFtBVrn1qJMT",
            "identifier": "C3nJhruVc7feyB6ckJwhi2",
            "operation": {{
                "data": {{
                    "attr_names": ["score"],
                    "name": "prefs",
                    "version": "1.0"
                }},
                "type": "101"
            }},
            "protocolVersion": 2,
            "reqId": 1613463373859595201,
            "signatures": {{
                "C3nJhruVc7feyB6ckJwhi2": {sigs[0]},
                "DJGEjaMunDtFtBVrn1qJMT": {sigs[1]}
            }}
        }}"""
        self.test_endorser_did = "DJGEjaMunDtFtBVrn1qJMT"
        self.test_endorser_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
        self.test_refuser_did = "AGDEjaMunDtFtBVrn1qPKQ"

        self.ledger = mock.create_autospec(BaseLedger)
        self.ledger.txn_endorse = mock.CoroutineMock(
            return_value=self.test_endorsed_message
        )
        self.ledger.register_nym = mock.CoroutineMock(return_value=(True, {}))

        self.context = AdminRequestContext.test_context({}, await create_test_profile())
        self.profile = self.context.profile
        injector = self.profile.context.injector
        injector.bind_instance(BaseLedger, self.ledger)
        injector.bind_instance(KeyTypes, KeyTypes())
        injector.bind_instance(DIDMethods, DIDMethods())

        async with self.profile.session() as session:
            self.wallet: BaseWallet = session.inject_or(BaseWallet)
            await self.wallet.create_local_did(
                SOV,
                ED25519,
                did="DJGEjaMunDtFtBVrn1qJMT",
                metadata={"meta": "data"},
            )
            await self.wallet.set_public_did("DJGEjaMunDtFtBVrn1qJMT")

        self.manager = TransactionManager(self.profile)

        assert self.manager.profile

    async def test_transaction_jobs(self):
        author = TransactionJob.TRANSACTION_AUTHOR
        endorser = TransactionJob.TRANSACTION_ENDORSER
        assert author != endorser

    async def test_create_record(self):
        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            transaction_record = await self.manager.create_record(
                messages_attach=self.test_messages_attach,
                connection_id=self.test_connection_id,
            )
            save_record.assert_called_once()

            assert (
                transaction_record.formats[0]["attach_id"]
                == transaction_record.messages_attach[0]["@id"]
            )
            assert (
                transaction_record.formats[0]["format"]
                == TransactionRecord.FORMAT_VERSION
            )
            assert (
                transaction_record.messages_attach[0]["data"]["json"]
                == self.test_messages_attach
            )
            assert transaction_record.state == TransactionRecord.STATE_TRANSACTION_CREATED

    async def test_txn_rec_retrieve_by_connection_and_thread_caching(self):
        async with self.profile.session() as sesn:
            sesn.context.injector.bind_instance(BaseCache, InMemoryCache())
            txn_rec = TransactionRecord(
                connection_id="123",
                thread_id="456",
            )
            await txn_rec.save(sesn)
            await TransactionRecord.retrieve_by_connection_and_thread(
                session=sesn,
                connection_id="123",
                thread_id="456",
            )  # set in cache
            await TransactionRecord.retrieve_by_connection_and_thread(
                session=sesn,
                connection_id="123",
                thread_id="456",
            )  # get from cache

    async def test_create_request_bad_state(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        transaction_record.state = TransactionRecord.STATE_TRANSACTION_ENDORSED

        with self.assertRaises(TransactionManagerError):
            await self.manager.create_request(transaction=transaction_record)

    async def test_create_request(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            (
                transaction_record,
                transaction_request,
            ) = await self.manager.create_request(
                transaction_record,
                expires_time=self.test_expires_time,
            )
            save_record.assert_called_once()

        assert transaction_record._type == TransactionRecord.SIGNATURE_REQUEST
        assert transaction_record.signature_request[0] == {
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signature_type": TransactionRecord.SIGNATURE_TYPE,
            "signer_goal_code": TransactionRecord.ENDORSE_TRANSACTION,
            "author_goal_code": TransactionRecord.WRITE_TRANSACTION,
        }
        assert transaction_record.state == TransactionRecord.STATE_REQUEST_SENT
        assert transaction_record.connection_id == self.test_connection_id
        assert transaction_record.timing["expires_time"] == self.test_expires_time

        assert transaction_request.transaction_id == transaction_record._id
        assert (
            transaction_request.signature_request
            == transaction_record.signature_request[0]
        )
        assert transaction_request.timing == transaction_record.timing
        assert (
            transaction_request.messages_attach == transaction_record.messages_attach[0]
        )

    async def test_create_request_author_did(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            (
                transaction_record,
                transaction_request,
            ) = await self.manager.create_request(
                transaction_record,
                expires_time=self.test_expires_time,
                author_goal_code=TransactionRecord.REGISTER_PUBLIC_DID,
                signer_goal_code=TransactionRecord.WRITE_DID_TRANSACTION,
            )
            save_record.assert_called_once()

        assert transaction_record._type == TransactionRecord.SIGNATURE_REQUEST
        assert transaction_record.signature_request[0] == {
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signature_type": TransactionRecord.SIGNATURE_TYPE,
            "signer_goal_code": TransactionRecord.WRITE_DID_TRANSACTION,
            "author_goal_code": TransactionRecord.REGISTER_PUBLIC_DID,
        }
        assert transaction_record.state == TransactionRecord.STATE_REQUEST_SENT
        assert transaction_record.connection_id == self.test_connection_id
        assert transaction_record.timing["expires_time"] == self.test_expires_time

        assert transaction_request.transaction_id == transaction_record._id
        assert (
            transaction_request.signature_request
            == transaction_record.signature_request[0]
        )
        assert transaction_request.timing == transaction_record.timing
        assert (
            transaction_request.messages_attach == transaction_record.messages_attach[0]
        )

    async def test_receive_request(self):
        mock_request = mock.MagicMock()
        mock_request.transaction_id = self.test_author_transaction_id
        mock_request.signature_request = {
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signature_type": TransactionRecord.SIGNATURE_TYPE,
            "signer_goal_code": TransactionRecord.ENDORSE_TRANSACTION,
            "author_goal_code": TransactionRecord.WRITE_TRANSACTION,
        }
        mock_request.messages_attach = {
            "@id": str(uuid4()),
            "mime-type": "application/json",
            "data": {"json": self.test_messages_attach},
        }
        mock_request.timing = {"expires_time": self.test_expires_time}

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            transaction_record = await self.manager.receive_request(
                mock_request, self.test_receivers_connection_id
            )
            save_record.assert_called_once()

        assert transaction_record._type == TransactionRecord.SIGNATURE_REQUEST
        assert transaction_record.signature_request[0] == mock_request.signature_request
        assert transaction_record.timing == mock_request.timing
        assert transaction_record.formats[0] == {
            "attach_id": mock_request.messages_attach["@id"],
            "format": TransactionRecord.FORMAT_VERSION,
        }
        assert transaction_record.messages_attach[0] == mock_request.messages_attach
        assert transaction_record.thread_id == self.test_author_transaction_id
        assert transaction_record.connection_id == self.test_receivers_connection_id
        assert transaction_record.state == TransactionRecord.STATE_REQUEST_RECEIVED

    async def test_create_endorse_response_bad_state(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        transaction_record.state = TransactionRecord.STATE_TRANSACTION_ENDORSED

        with self.assertRaises(TransactionManagerError):
            await self.manager.create_endorse_response(
                transaction=transaction_record,
                state=TransactionRecord.STATE_TRANSACTION_ENDORSED,
            )

    async def test_create_endorse_response(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        transaction_record.state = TransactionRecord.STATE_REQUEST_RECEIVED
        transaction_record.thread_id = self.test_author_transaction_id

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            (
                transaction_record,
                endorsed_transaction_response,
            ) = await self.manager.create_endorse_response(
                transaction_record,
                state=TransactionRecord.STATE_TRANSACTION_ENDORSED,
            )
            save_record.assert_called_once()

        assert transaction_record._type == TransactionRecord.SIGNATURE_RESPONSE
        assert (
            transaction_record.messages_attach[0]["data"]["json"]
            == self.test_endorsed_message
        )
        assert transaction_record.signature_response[0] == {
            "message_id": transaction_record.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": TransactionRecord.ENDORSE_TRANSACTION,
            "signature_type": TransactionRecord.SIGNATURE_TYPE,
            "signature": {self.test_endorser_did: self.test_signature},
        }
        assert transaction_record.state == TransactionRecord.STATE_TRANSACTION_ENDORSED

        assert (
            endorsed_transaction_response.transaction_id
            == self.test_author_transaction_id
        )
        assert endorsed_transaction_response.thread_id == transaction_record._id
        assert endorsed_transaction_response.signature_response == {
            "message_id": transaction_record.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": TransactionRecord.ENDORSE_TRANSACTION,
            "signature_type": TransactionRecord.SIGNATURE_TYPE,
            "signature": {self.test_endorser_did: self.test_signature},
        }

        assert (
            endorsed_transaction_response.state
            == TransactionRecord.STATE_TRANSACTION_ENDORSED
        )
        assert endorsed_transaction_response.endorser_did == self.test_endorser_did

    async def test_create_endorse_response_author_did(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            (
                transaction_record,
                transaction_request,
            ) = await self.manager.create_request(
                transaction_record,
                expires_time=self.test_expires_time,
                author_goal_code=TransactionRecord.REGISTER_PUBLIC_DID,
                signer_goal_code=TransactionRecord.WRITE_DID_TRANSACTION,
            )
            save_record.assert_called_once()

        transaction_record.state = TransactionRecord.STATE_REQUEST_RECEIVED
        transaction_record.thread_id = self.test_author_transaction_id
        transaction_record.messages_attach[0]["data"]["json"] = json.dumps(
            {
                "did": "test",
                "verkey": "test",
                "alias": "test",
                "role": "",
            }
        )

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            (
                transaction_record,
                endorsed_transaction_response,
            ) = await self.manager.create_endorse_response(
                transaction_record,
                state=TransactionRecord.STATE_TRANSACTION_ENDORSED,
            )
            save_record.assert_called_once()

        assert transaction_record._type == TransactionRecord.SIGNATURE_RESPONSE
        assert (
            transaction_record.messages_attach[0]["data"]["json"]
            == '{"result": {"txn": {"type": "1", "data": {"dest": "test"}}}, "meta_data": {"did": "test", "verkey": "test", "alias": "test", "role": ""}}'
        )

    async def test_receive_endorse_response(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        self.test_author_transaction_id = transaction_record._id

        mock_response = mock.MagicMock()
        mock_response.transaction_id = self.test_author_transaction_id
        mock_response.thread_id = self.test_endorser_transaction_id
        mock_response.signature_response = {
            "message_id": transaction_record.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": TransactionRecord.ENDORSE_TRANSACTION,
            "signature_type": TransactionRecord.SIGNATURE_TYPE,
            "signature": {self.test_endorser_did: self.test_signature},
        }
        mock_response.state = TransactionRecord.STATE_TRANSACTION_ENDORSED
        mock_response.endorser_did = self.test_endorser_did

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            transaction_record = await self.manager.receive_endorse_response(
                mock_response
            )
            save_record.assert_called_once()

        assert transaction_record._type == TransactionRecord.SIGNATURE_RESPONSE
        assert transaction_record.state == TransactionRecord.STATE_TRANSACTION_ENDORSED
        assert transaction_record.signature_response[0] == {
            "message_id": transaction_record.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": TransactionRecord.ENDORSE_TRANSACTION,
            "signature_type": TransactionRecord.SIGNATURE_TYPE,
            "signature": {self.test_endorser_did: self.test_signature},
        }
        assert transaction_record.thread_id == self.test_endorser_transaction_id
        assert (
            transaction_record.messages_attach[0]["data"]["json"] == self.test_signature
        )

    async def test_complete_transaction(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        future = asyncio.Future()
        future.set_result(
            mock.MagicMock(return_value=mock.MagicMock(add_record=mock.CoroutineMock()))
        )
        self.ledger.get_indy_storage = future
        self.ledger.txn_submit = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "result": {
                        "txn": {"type": "101", "metadata": {"from": TEST_DID}},
                        "txnMetadata": {"txnId": SCHEMA_ID},
                    }
                }
            )
        )

        with (
            mock.patch.object(TransactionRecord, "save", autospec=True) as save_record,
            mock.patch.object(ConnRecord, "retrieve_by_id") as mock_conn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_their_job": (
                            TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                        "transaction_my_job": (TransactionJob.TRANSACTION_AUTHOR.name),
                    }
                )
            )

            (
                transaction_record,
                transaction_acknowledgement_message,
            ) = await self.manager.complete_transaction(transaction_record, False)
            save_record.assert_called_once()

        assert transaction_record.state == TransactionRecord.STATE_TRANSACTION_ACKED

    @mock.patch.object(
        LegacyIndyRegistry,
        "txn_submit",
        return_value=json.dumps(
            {
                "result": {
                    "txn": {"type": "101", "metadata": {"from": TEST_DID}},
                    "txnMetadata": {"txnId": SCHEMA_ID},
                },
            }
        ),
    )
    @mock.patch.object(AnonCredsIssuer, "finish_schema")
    async def test_complete_transaction_anoncreds(
        self, mock_finish_schema, mock_txn_submit
    ):
        self.profile.settings.set_value("wallet.type", "askar-anoncreds")

        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
            meta_data={
                "context": {
                    "job_id": "217544da8ab14b12b18eccd11f07d269",
                    "schema_id": "FB5yHWKaZk59hiKqjJKEHs:2:author-schema:3.3",
                }
            },
        )
        future = asyncio.Future()
        future.set_result(
            mock.MagicMock(return_value=mock.MagicMock(add_record=mock.CoroutineMock()))
        )
        self.ledger.get_indy_storage = future

        with (
            mock.patch.object(TransactionRecord, "save", autospec=True) as save_record,
            mock.patch.object(ConnRecord, "retrieve_by_id") as mock_conn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_their_job": (
                            TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                        "transaction_my_job": (TransactionJob.TRANSACTION_AUTHOR.name),
                    }
                )
            )

            (
                transaction_record,
                transaction_acknowledgement_message,
            ) = await self.manager.complete_transaction(transaction_record, False)
            save_record.assert_called_once()

        assert transaction_record.state == TransactionRecord.STATE_TRANSACTION_ACKED
        assert mock_txn_submit.called
        assert mock_finish_schema.called

    async def test_create_refuse_response_bad_state(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        transaction_record.state = TransactionRecord.STATE_TRANSACTION_ENDORSED

        with self.assertRaises(TransactionManagerError):
            await self.manager.create_refuse_response(
                transaction=transaction_record,
                state=TransactionRecord.STATE_TRANSACTION_REFUSED,
                refuser_did=self.test_refuser_did,
            )

    async def test_create_refuse_response(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        transaction_record.state = TransactionRecord.STATE_REQUEST_RECEIVED
        transaction_record.thread_id = self.test_author_transaction_id

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            (
                transaction_record,
                refused_transaction_response,
            ) = await self.manager.create_refuse_response(
                transaction_record,
                state=TransactionRecord.STATE_TRANSACTION_REFUSED,
                refuser_did=self.test_refuser_did,
            )
            save_record.assert_called_once()

        assert transaction_record._type == TransactionRecord.SIGNATURE_RESPONSE
        assert transaction_record.signature_response[0] == {
            "message_id": transaction_record.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": TransactionRecord.REFUSE_TRANSACTION,
        }
        assert transaction_record.state == TransactionRecord.STATE_TRANSACTION_REFUSED

        assert (
            refused_transaction_response.transaction_id == self.test_author_transaction_id
        )
        assert refused_transaction_response.thread_id == transaction_record._id
        assert refused_transaction_response.signature_response == {
            "message_id": transaction_record.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": TransactionRecord.REFUSE_TRANSACTION,
        }

        assert (
            refused_transaction_response.state
            == TransactionRecord.STATE_TRANSACTION_REFUSED
        )
        assert refused_transaction_response.endorser_did == self.test_refuser_did

    async def test_receive_refuse_response(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        self.test_author_transaction_id = transaction_record._id

        mock_response = mock.MagicMock()
        mock_response.transaction_id = self.test_author_transaction_id
        mock_response.thread_id = self.test_endorser_transaction_id
        mock_response.signature_response = {
            "message_id": transaction_record.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": TransactionRecord.REFUSE_TRANSACTION,
        }
        mock_response.state = TransactionRecord.STATE_TRANSACTION_REFUSED
        mock_response.endorser_did = self.test_refuser_did

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            transaction_record = await self.manager.receive_refuse_response(mock_response)
            save_record.assert_called_once()

        assert transaction_record._type == TransactionRecord.SIGNATURE_RESPONSE
        assert transaction_record.state == TransactionRecord.STATE_TRANSACTION_REFUSED
        assert transaction_record.signature_response[0] == {
            "message_id": transaction_record.messages_attach[0]["@id"],
            "context": TransactionRecord.SIGNATURE_CONTEXT,
            "method": TransactionRecord.ADD_SIGNATURE,
            "signer_goal_code": TransactionRecord.REFUSE_TRANSACTION,
        }
        assert transaction_record.thread_id == self.test_endorser_transaction_id

    async def test_cancel_transaction_bad_state(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        transaction_record.state = TransactionRecord.STATE_TRANSACTION_ENDORSED

        with self.assertRaises(TransactionManagerError):
            await self.manager.cancel_transaction(
                transaction=transaction_record,
                state=TransactionRecord.STATE_TRANSACTION_CANCELLED,
            )

    async def test_cancel_transaction(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        transaction_record.state = TransactionRecord.STATE_REQUEST_SENT
        transaction_record.thread_id = self.test_endorser_transaction_id
        transaction_record._id = self.test_author_transaction_id

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            (
                transaction_record,
                cancelled_transaction_response,
            ) = await self.manager.cancel_transaction(
                transaction_record, state=TransactionRecord.STATE_TRANSACTION_CANCELLED
            )
            save_record.assert_called_once()

        assert transaction_record.state == TransactionRecord.STATE_TRANSACTION_CANCELLED

        assert cancelled_transaction_response.thread_id == self.test_author_transaction_id
        assert (
            cancelled_transaction_response.state
            == TransactionRecord.STATE_TRANSACTION_CANCELLED
        )

    async def test_receive_cancel_transaction(self):
        author_transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        (
            author_transaction_record,
            author_transaction_request,
        ) = await self.manager.create_request(author_transaction_record)

        endorser_transaction_record = await self.manager.receive_request(
            author_transaction_request, self.test_receivers_connection_id
        )

        mock_response = mock.MagicMock()
        mock_response.state = TransactionRecord.STATE_TRANSACTION_CANCELLED
        mock_response.thread_id = author_transaction_record._id

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            endorser_transaction_record = await self.manager.receive_cancel_transaction(
                mock_response, self.test_receivers_connection_id
            )
            save_record.assert_called_once()

        assert (
            endorser_transaction_record.state
            == TransactionRecord.STATE_TRANSACTION_CANCELLED
        )

    async def test_transaction_resend_bad_state(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        transaction_record.state = TransactionRecord.STATE_TRANSACTION_ENDORSED

        with self.assertRaises(TransactionManagerError):
            await self.manager.transaction_resend(
                transaction=transaction_record,
                state=TransactionRecord.STATE_TRANSACTION_RESENT,
            )

    async def test_transaction_resend(self):
        transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        transaction_record.state = TransactionRecord.STATE_TRANSACTION_REFUSED
        transaction_record.thread_id = self.test_endorser_transaction_id
        transaction_record._id = self.test_author_transaction_id

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            (
                transaction_record,
                resend_transaction_response,
            ) = await self.manager.transaction_resend(
                transaction_record, state=TransactionRecord.STATE_TRANSACTION_RESENT
            )
            save_record.assert_called_once()

        assert transaction_record.state == TransactionRecord.STATE_TRANSACTION_RESENT

        assert resend_transaction_response.thread_id == self.test_author_transaction_id
        assert (
            resend_transaction_response.state
            == TransactionRecord.STATE_TRANSACTION_RESENT_RECEIVED
        )

    async def test_receive_transaction_resend(self):
        author_transaction_record = await self.manager.create_record(
            messages_attach=self.test_messages_attach,
            connection_id=self.test_connection_id,
        )
        (
            author_transaction_record,
            author_transaction_request,
        ) = await self.manager.create_request(author_transaction_record)

        endorser_transaction_record = await self.manager.receive_request(
            author_transaction_request, self.test_receivers_connection_id
        )

        mock_response = mock.MagicMock()
        mock_response.state = TransactionRecord.STATE_TRANSACTION_RESENT_RECEIVED
        mock_response.thread_id = author_transaction_record._id

        with mock.patch.object(TransactionRecord, "save", autospec=True) as save_record:
            endorser_transaction_record = await self.manager.receive_transaction_resend(
                mock_response, self.test_receivers_connection_id
            )
            save_record.assert_called_once()

        assert (
            endorser_transaction_record.state
            == TransactionRecord.STATE_TRANSACTION_RESENT_RECEIVED
        )

    async def test_set_transaction_my_job(self):
        conn_record = mock.MagicMock(
            metadata_get=mock.CoroutineMock(
                side_effect=[
                    None,
                    {"meta": "data"},
                ]
            ),
            metadata_set=mock.CoroutineMock(),
        )

        for i in range(2):
            await self.manager.set_transaction_my_job(conn_record, "Hello")

    async def test_set_transaction_their_job(self):
        mock_job = mock.MagicMock()
        mock_conn = mock.MagicMock()
        mock_conn.metadata_get = mock.CoroutineMock(
            side_effect=[
                None,
                {"meta": "data"},
            ]
        )
        mock_conn.metadata_set = mock.CoroutineMock()

        for i in range(2):
            await self.manager.set_transaction_their_job(mock_job, mock_conn)

    @mock.patch.object(AnonCredsIssuer, "finish_schema")
    @mock.patch.object(AnonCredsIssuer, "finish_cred_def")
    @mock.patch.object(AnonCredsRevocation, "finish_revocation_registry_definition")
    @mock.patch.object(AnonCredsRevocation, "finish_revocation_list")
    async def test_endorsed_txn_post_processing_anoncreds(
        self,
        mock_finish_revocation_list,
        mock_finish_revocation_registry_definition,
        mock_finish_cred_def,
        mock_finish_schema,
    ):
        self.profile.settings.set_value("wallet.type", "askar-anoncreds")
        transaction = TransactionRecord(
            connection_id="123",
            thread_id="456",
            meta_data={
                "context": {
                    "job_id": "217544da8ab14b12b18eccd11f07d269",
                    "schema_id": SCHEMA_ID,
                }
            },
        )

        ledger_response = {
            "result": {
                "txn": {"type": "101", "metadata": {"from": TEST_DID}},
                "txnMetadata": {"txnId": SCHEMA_ID},
            },
        }

        await self.manager.endorsed_txn_post_processing(transaction, ledger_response)
        mock_finish_schema.assert_called_once()

        transaction = TransactionRecord(
            connection_id="123",
            thread_id="456",
            meta_data={
                "context": {
                    "job_id": "217544da8ab14b12b18eccd11f07d269",
                    "cred_def_id": CRED_DEF_ID,
                    "issuer_did": "TUku9MDGa7QALbAJX4oAww",
                    "options": {},
                }
            },
        )

        ledger_response = {
            "result": {
                "txn": {
                    "type": "102",
                    "metadata": {"from": TEST_DID},
                    "data": {"ref": 1},
                },
                "txnMetadata": {"txnId": CRED_DEF_ID},
            },
        }

        await self.manager.endorsed_txn_post_processing(transaction, ledger_response)
        mock_finish_cred_def.assert_called_once()

        transaction = TransactionRecord(
            connection_id="123",
            thread_id="456",
            meta_data={
                "context": {
                    "job_id": "217544da8ab14b12b18eccd11f07d269",
                    "rev_reg_id": REV_REG_ID,
                    "options": {},
                }
            },
        )

        ledger_response = {
            "result": {
                "txn": {
                    "type": "113",
                    "metadata": {"from": TEST_DID},
                },
                "txnMetadata": {"txnId": REV_REG_ID},
            },
        }

        await self.manager.endorsed_txn_post_processing(transaction, ledger_response)
        mock_finish_revocation_registry_definition.assert_called_once()

        transaction = TransactionRecord(
            connection_id="123",
            thread_id="456",
            meta_data={
                "context": {
                    "job_id": "217544da8ab14b12b18eccd11f07d269",
                    "rev_reg_id": REV_REG_ID,
                }
            },
        )

        ledger_response = {
            "result": {
                "txn": {
                    "type": "114",
                    "metadata": {"from": TEST_DID},
                    "data": {"revocRegDefId": REV_REG_ID, "value": {"revoked": [1]}},
                },
                "txnMetadata": {"txnId": REV_REG_ID},
            },
        }

        await self.manager.endorsed_txn_post_processing(transaction, ledger_response)
        mock_finish_revocation_list.assert_called_once()



"""Help command for indexing available commands."""

from typing import Sequence

from configargparse import ArgumentParser

from ..version import __version__


def execute(argv: Sequence[str] = None):
    """Execute the help command."""
    from . import PROG, available_commands, load_command

    parser = ArgumentParser(prog=PROG)
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="print application version and exit",
    )
    subparsers = parser.add_subparsers()
    for cmd in available_commands():
        if cmd["name"] == "help":
            continue
        module = load_command(cmd["name"])
        subparser = subparsers.add_parser(cmd["name"], help=cmd["summary"])
        module.init_argument_parser(subparser)
    args = parser.parse_args(argv)
    if args.version:
        print(__version__)
    else:
        parser.print_help()


def main():
    """Execute the main line."""
    if __name__ == "__main__":
        execute()


main()



"""Test Problem Report Message."""

from unittest import TestCase, mock

import pytest

from ......messaging.models.base import BaseModelError
from .. import problem_report as test_module
from ..problem_report import (
    OOBProblemReport,
    OOBProblemReportSchema,
    ProblemReportReason,
    ValidationError,
)


class TestOOBProblemReportMessage(TestCase):
    """Test problem report."""

    def setUp(self):
        self.problem_report = OOBProblemReport(
            description={
                "en": "Test",
                "code": ProblemReportReason.EXISTING_CONNECTION_NOT_ACTIVE.value,
            }
        )

    def test_init(self):
        """Test initialization."""
        self.problem_report.assign_thread_id(thid="test_thid", pthid="test_pthid")
        assert isinstance(self.problem_report, OOBProblemReport)
        assert isinstance(self.problem_report._id, str)
        assert len(self.problem_report._id) > 4
        assert self.problem_report._thread.thid == "test_thid"
        assert self.problem_report._thread.pthid == "test_pthid"

    def test_make_model(self):
        """Make problem report model."""
        self.problem_report.assign_thread_id(thid="test_thid", pthid="test_pthid")
        data = self.problem_report.serialize()
        model_instance = OOBProblemReport.deserialize(data)
        assert isinstance(model_instance, OOBProblemReport)

        model_instance.description["code"] = "extraneous code"
        with pytest.raises(BaseModelError):
            OOBProblemReport.deserialize(model_instance)

    def test_pre_dump_x(self):
        """Exercise pre-dump serialization requirements."""
        with pytest.raises(BaseModelError):
            data = self.problem_report.serialize()

    def test_validate_x(self):
        """Exercise validation requirements."""
        schema = OOBProblemReportSchema()
        with pytest.raises(ValidationError):
            schema.validate_fields({})

    def test_validate_and_logger(self):
        """Capture ValidationError and Logs."""
        data = OOBProblemReport(
            description={
                "en": "Insufficient credit",
                "code": "invalid_code",
            },
        )
        data.assign_thread_id(thid="test_thid", pthid="test_pthid")
        data = data.serialize()
        with mock.patch.object(test_module, "LOGGER", autospec=True) as mock_logger:
            OOBProblemReportSchema().validate_fields(data)
        assert mock_logger.warning.call_count == 1

    def test_assign_msg_type_version_to_model_inst(self):
        test_msg = OOBProblemReport()
        assert "1.1" in test_msg._type
        assert "1.1" in OOBProblemReport.Meta.message_type
        test_msg = OOBProblemReport(version="1.2")
        assert "1.2" in test_msg._type
        assert "1.1" in OOBProblemReport.Meta.message_type



"""BbsBlsSignature2020 class."""

from datetime import datetime
from typing import List, Optional, Union

from dateutil import tz

from ....wallet.util import b64_to_bytes, bytes_to_b64
from ..crypto import _KeyPair as KeyPair
from ..document_loader import DocumentLoaderMethod
from ..error import LinkedDataProofException
from ..purposes import _ProofPurpose as ProofPurpose
from ..validation_result import ProofResult
from .bbs_bls_signature_2020_base import BbsBlsSignature2020Base


class BbsBlsSignature2020(BbsBlsSignature2020Base):
    """BbsBlsSignature2020 class."""

    signature_type = "BbsBlsSignature2020"

    def __init__(
        self,
        *,
        key_pair: KeyPair,
        proof: Optional[dict] = None,
        verification_method: Optional[str] = None,
        date: Union[datetime, None] = None,
    ):
        """Create new BbsBlsSignature2020 instance.

        Args:
            key_pair (KeyPair): Key pair to use. Must provide BBS signatures
            proof (dict, optional): A JSON-LD document with options to use for the
                `proof` node (e.g. any other custom fields can be provided here
                using a context different from security-v2).
            verification_method (str, optional): A key id URL to the paired public key.
            date (datetime, optional): Signing date to use. Defaults to now

        """
        super().__init__(proof=proof)
        self.key_pair = key_pair
        self.verification_method = verification_method
        self.date = date

    async def create_proof(
        self,
        *,
        document: dict,
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> dict:
        """Create proof for document, return proof."""
        proof = self.proof.copy() if self.proof else {}

        proof["type"] = self.signature_type
        proof["verificationMethod"] = self.verification_method

        # Set created if not already set
        if not proof.get("created"):
            # Use class date, or now
            date = self.date or datetime.now(tz.UTC)
            if not date.tzinfo:
                date = date.replace(tzinfo=tz.UTC)
            proof["created"] = date.isoformat(timespec="seconds")

        # Allow purpose to update the proof; the `proof` is in the
        # SECURITY_CONTEXT_URL `@context` -- therefore the `purpose` must
        # ensure any added fields are also represented in that same `@context`
        proof = purpose.update(proof)

        # Create statements to sign
        verify_data = self._create_verify_data(
            proof=proof, document=document, document_loader=document_loader
        )

        # Encode statements as bytes
        verify_data = [item.encode("utf-8") for item in verify_data]

        # Sign statements
        proof = await self.sign(verify_data=verify_data, proof=proof)

        return proof

    async def verify_proof(
        self,
        *,
        proof: dict,
        document: dict,
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> ProofResult:
        """Verify proof against document and proof purpose."""
        try:
            # Create statements to verify
            verify_data = self._create_verify_data(
                proof=proof, document=document, document_loader=document_loader
            )

            # Encode statements as bytes
            verify_data = [item.encode("utf-8") for item in verify_data]

            # Fetch verification method
            verification_method = self._get_verification_method(
                proof=proof, document_loader=document_loader
            )

            # Verify signature on data
            verified = await self.verify_signature(
                verify_data=verify_data,
                verification_method=verification_method,
                document=document,
                proof=proof,
                document_loader=document_loader,
            )
            if not verified:
                raise LinkedDataProofException(
                    f"Invalid signature on document {document}"
                )

            # Ensure proof was performed for a valid purpose
            purpose_result = purpose.validate(
                proof=proof,
                document=document,
                suite=self,
                verification_method=verification_method,
                document_loader=document_loader,
            )

            if not purpose_result.valid:
                return ProofResult(
                    verified=False,
                    purpose_result=purpose_result,
                    error=purpose_result.error,
                )

            return ProofResult(verified=True, purpose_result=purpose_result)
        except Exception as err:
            return ProofResult(verified=False, error=err)

    def _create_verify_data(
        self, *, proof: dict, document: dict, document_loader: DocumentLoaderMethod
    ) -> List[str]:
        """Create verification data.

        Returns a list of canonized statements

        """
        proof_statements = self._create_verify_proof_data(
            proof=proof, document=document, document_loader=document_loader
        )
        document_statements = self._create_verify_document_data(
            document=document, document_loader=document_loader
        )

        return [*proof_statements, *document_statements]

    def _canonize_proof(
        self, *, proof: dict, document: dict, document_loader: DocumentLoaderMethod
    ):
        """Canonize proof dictionary. Removes value that are not part of signature."""
        # Use default security context url if document has no context
        proof = {**proof, "@context": document.get("@context") or self._default_proof}

        proof.pop("proofValue", None)

        return self._canonize(input=proof, document_loader=document_loader)

    async def sign(self, *, verify_data: List[bytes], proof: dict) -> dict:
        """Sign the data and add it to the proof.

        Args:
            verify_data (List[bytes]): The data to sign.
            proof (dict): The proof to add the signature to

        Returns:
            dict: The proof object with the added signature

        """
        signature = await self.key_pair.sign(verify_data)

        proof["proofValue"] = bytes_to_b64(
            signature, urlsafe=False, pad=True, encoding="utf-8"
        )

        return proof

    async def verify_signature(
        self,
        *,
        verify_data: List[bytes],
        verification_method: dict,
        document: dict,
        proof: dict,
        document_loader: DocumentLoaderMethod,
    ) -> bool:
        """Verify the data against the proof.

        Args:
            verify_data (bytes): The data to check
            verification_method (dict): The verification method to use.
            document (dict): The document the verify data is derived for as extra context
            proof (dict): The proof to check
            document_loader (DocumentLoader): Document loader used for resolving

        Returns:
            bool: Whether the signature is valid for the data

        """

        if not (isinstance(proof.get("proofValue"), str)):
            raise LinkedDataProofException(
                'The proof does not contain a valid "proofValue" property.'
            )

        signature = b64_to_bytes(proof["proofValue"])

        # If the key pair has no public key yet, create a new key pair
        # from the verification method. We don't want to overwrite data
        # on the original key pair
        key_pair = self.key_pair
        if not key_pair.has_public_key:
            key_pair = key_pair.from_verification_method(verification_method)

        return await key_pair.verify(verify_data, signature)



"""Utilities to deal with anoncreds objects."""

from ..holder import AnonCredsHolder


def _get_value_error_msg(proof_request: dict, referent: str) -> str:
    return (
        "Could not automatically construct presentation for "
        + f"presentation request {proof_request['name']}"
        + f":{proof_request['version']} because referent "
        + f"{referent} did not produce any credentials."
    )


async def get_requested_creds_from_proof_request_preview(
    proof_request: dict,
    *,
    holder: AnonCredsHolder,
) -> dict[str, dict]:
    """Build anoncreds requested-credentials structure.

    Given input proof request and presentation preview, use credentials in
    holder's wallet to build anoncreds requested credentials structure for input
    to proof creation.

    Args:
        proof_request: anoncreds proof request
        preview: preview from presentation proposal, if applicable
        holder: holder injected into current context

    """
    req_creds = {
        "self_attested_attributes": {},
        "requested_attributes": {},
        "requested_predicates": {},
    }

    for referent, _ in proof_request["requested_attributes"].items():
        credentials = await holder.get_credentials_for_presentation_request_by_referent(
            presentation_request=proof_request,
            referents=(referent,),
            offset=0,
            limit=100,
        )
        if not credentials:
            raise ValueError(_get_value_error_msg(proof_request, referent))

        cred_match = credentials[0]  # holder sorts

        if "restrictions" in proof_request["requested_attributes"][referent]:
            req_creds["requested_attributes"][referent] = {
                "cred_id": cred_match["cred_info"]["referent"],
                "revealed": True,
            }
        else:
            req_creds["self_attested_attributes"][referent] = cred_match["cred_info"][
                "attrs"
            ][proof_request["requested_attributes"][referent]["name"]]

    for referent in proof_request["requested_predicates"]:
        credentials = await holder.get_credentials_for_presentation_request_by_referent(
            presentation_request=proof_request,
            referents=(referent,),
            offset=0,
            limit=100,
        )
        if not credentials:
            raise ValueError(_get_value_error_msg(proof_request, referent))

        cred_match = credentials[0]  # holder sorts
        if "restrictions" in proof_request["requested_predicates"][referent]:
            req_creds["requested_predicates"][referent] = {
                "cred_id": cred_match["cred_info"]["referent"],
                "revealed": True,
            }
        else:
            req_creds["self_attested_attributes"][referent] = cred_match["cred_info"][
                "attrs"
            ][proof_request["requested_predicates"][referent]["name"]]

    return req_creds


def extract_non_revocation_intervals_from_proof_request(proof_req: dict) -> dict:
    """Return non-revocation intervals by requested item referent in proof request."""
    non_revoc_intervals = {}
    for req_item_type in ("requested_attributes", "requested_predicates"):
        for reft, req_item in proof_req[req_item_type].items():
            interval = req_item.get(
                "non_revoked",
                proof_req.get("non_revoked"),
            )
            if interval:
                timestamp_from = interval.get("from")
                timestamp_to = interval.get("to")
                if (timestamp_to is not None) and timestamp_from == timestamp_to:
                    interval["from"] = 0  # accommodate verify=False if from=to
            non_revoc_intervals[reft] = interval
    return non_revoc_intervals



"""Test LegacyPeerDIDResolver."""

import pydid
import pytest
import pytest_asyncio

from ....cache.base import BaseCache
from ....cache.in_memory import InMemoryCache
from ....connections.models.diddoc.diddoc import DIDDoc
from ....core.profile import Profile
from ....storage.error import StorageNotFoundError
from ....tests import mock
from ....utils.testing import create_test_profile
from .. import legacy_peer as test_module
from ..legacy_peer import LegacyPeerDIDResolver

TEST_DID0 = "did:sov:WgWxqztrNooG92RXvxSTWv"
TEST_DID1 = "did:example:abc123"
TEST_DID2 = "did:sov:5No7f9KvpsfSqd6xsGxABh"


@pytest.fixture
def resolver():
    """Resolver fixture."""
    yield LegacyPeerDIDResolver()


@pytest_asyncio.fixture
async def profile():
    """Profile fixture."""
    profile = await create_test_profile()
    profile.context.injector.bind_instance(BaseCache, InMemoryCache())
    yield profile


class TestLegacyPeerDIDResolver:
    @pytest.mark.asyncio
    async def test_supports(self, resolver: LegacyPeerDIDResolver, profile: Profile):
        """Test supports."""
        with mock.patch.object(test_module, "BaseConnectionManager") as mock_mgr:
            mock_mgr.return_value = mock.MagicMock(
                fetch_did_document=mock.CoroutineMock(
                    return_value=(DIDDoc(TEST_DID0), None)
                )
            )
            assert await resolver.supports(profile, TEST_DID0)

    @pytest.mark.asyncio
    async def test_supports_no_cache(
        self, resolver: LegacyPeerDIDResolver, profile: Profile
    ):
        """Test supports."""
        profile.context.injector.clear_binding(BaseCache)
        with mock.patch.object(test_module, "BaseConnectionManager") as mock_mgr:
            mock_mgr.return_value = mock.MagicMock(
                fetch_did_document=mock.CoroutineMock(
                    return_value=(DIDDoc(TEST_DID0), None)
                )
            )
            assert await resolver.supports(profile, TEST_DID0)

    @pytest.mark.asyncio
    async def test_supports_x_not_legacy_did(
        self, resolver: LegacyPeerDIDResolver, profile: Profile
    ):
        """Test supports returns false for DID not matching legacy."""
        assert not await resolver.supports(profile, TEST_DID1)

    @pytest.mark.asyncio
    async def test_supports_x_unknown_did(
        self, resolver: LegacyPeerDIDResolver, profile: Profile
    ):
        """Test supports returns false for unknown DID."""
        with mock.patch.object(test_module, "BaseConnectionManager") as mock_mgr:
            mock_mgr.return_value = mock.MagicMock(
                fetch_did_document=mock.CoroutineMock(side_effect=StorageNotFoundError)
            )
            assert not await resolver.supports(profile, TEST_DID2)

    @pytest.mark.asyncio
    async def test_resolve(self, resolver: LegacyPeerDIDResolver, profile: Profile):
        """Test resolve."""
        with (
            mock.patch.object(test_module, "BaseConnectionManager") as mock_mgr,
            mock.patch.object(test_module, "LegacyDocCorrections") as mock_corrections,
        ):
            doc = object()
            mock_corrections.apply = mock.MagicMock(return_value=doc)
            mock_mgr.return_value = mock.MagicMock(
                fetch_did_document=mock.CoroutineMock(
                    return_value=(DIDDoc(TEST_DID0), None)
                )
            )
            result = await resolver.resolve(profile, TEST_DID0)
            assert result == doc

    @pytest.mark.asyncio
    async def test_resolve_x_not_found(
        self, resolver: LegacyPeerDIDResolver, profile: Profile
    ):
        """Test resolve with not found.

        This should be impossible in practice but still.
        """
        with (
            mock.patch.object(test_module, "BaseConnectionManager") as mock_mgr,
            mock.patch.object(test_module, "LegacyDocCorrections") as mock_corrections,
            pytest.raises(test_module.DIDNotFound),
        ):
            doc = object
            mock_corrections.apply = mock.MagicMock(return_value=doc)
            mock_mgr.return_value = mock.MagicMock(
                fetch_did_document=mock.CoroutineMock(side_effect=StorageNotFoundError)
            )
            resolver.supports = mock.CoroutineMock(return_value=True)
            await resolver.resolve(profile, TEST_DID0)

    @pytest.mark.parametrize(
        ("input_doc", "expected"),
        [
            (  # Examples
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                    "publicKey": [
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                            "publicKeyBase58": "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG",
                        }
                    ],
                    "authentication": [
                        {
                            "type": "Ed25519SignatureAuthentication2018",
                            "publicKey": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1",
                        }
                    ],
                    "service": [
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ;indy",
                            "type": "IndyAgent",
                            "priority": 0,
                            "recipientKeys": [
                                "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG"
                            ],
                            "routingKeys": [
                                "9NnKFUZoYcCqYC2PcaXH3cnaGsoRfyGgyEHbvbLJYh8j"
                            ],
                            "serviceEndpoint": "http://bob:3000",
                        }
                    ],
                },
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                    "verificationMethod": [
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                            "publicKeyBase58": "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG",
                        }
                    ],
                    "authentication": ["did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1"],
                    "service": [
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#didcomm-0",
                            "type": "did-communication",
                            "priority": 0,
                            "recipientKeys": ["did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1"],
                            "routingKeys": [
                                "did:key:z6Mknq3MqipEt9hJegs6J9V7tiLa6T5H5rX3fFCXksJKTuv7#z6Mknq3MqipEt9hJegs6J9V7tiLa6T5H5rX3fFCXksJKTuv7"
                            ],
                            "serviceEndpoint": "http://bob:3000",
                        }
                    ],
                },
            ),
            (  # Stored findy doc
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:5qXMeLdyWEQhieFUBNw5ux",
                    "publicKey": [
                        {
                            "id": "did:sov:5qXMeLdyWEQhieFUBNw5ux#1",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:5qXMeLdyWEQhieFUBNw5ux",
                            "publicKeyBase58": "3dtu2WWtd5ELwRTJEPzmEJUYEp8Qq36N2QA24g9tFXK9",
                        }
                    ],
                    "authentication": [
                        {
                            "type": "Ed25519SignatureAuthentication2018",
                            "publicKey": "did:sov:5qXMeLdyWEQhieFUBNw5ux#1",
                        }
                    ],
                    "service": [
                        {
                            "id": "did:sov:5qXMeLdyWEQhieFUBNw5ux",
                            "type": "IndyAgent",
                            "priority": None,
                            "recipientKeys": [
                                "3dtu2WWtd5ELwRTJEPzmEJUYEp8Qq36N2QA24g9tFXK9"
                            ],
                            "serviceEndpoint": "http://172.17.0.1:9031/a2a/5b6dyY6PndLaCnWxZbeEYW/5b6dyY6PndLaCnWxZbeEYW/2f6aae0c-6b04-40ff-a25e-faecaea39f83",
                        }
                    ],
                },
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:5qXMeLdyWEQhieFUBNw5ux",
                    "verificationMethod": [
                        {
                            "id": "did:sov:5qXMeLdyWEQhieFUBNw5ux#1",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:5qXMeLdyWEQhieFUBNw5ux",
                            "publicKeyBase58": "3dtu2WWtd5ELwRTJEPzmEJUYEp8Qq36N2QA24g9tFXK9",
                        }
                    ],
                    "authentication": ["did:sov:5qXMeLdyWEQhieFUBNw5ux#1"],
                    "service": [
                        {
                            "id": "did:sov:5qXMeLdyWEQhieFUBNw5ux#didcomm-0",
                            "type": "did-communication",
                            "recipientKeys": ["did:sov:5qXMeLdyWEQhieFUBNw5ux#1"],
                            "serviceEndpoint": "http://172.17.0.1:9031/a2a/5b6dyY6PndLaCnWxZbeEYW/5b6dyY6PndLaCnWxZbeEYW/2f6aae0c-6b04-40ff-a25e-faecaea39f83",
                        }
                    ],
                },
            ),
            (  # Stored afgo doc
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:1H6d1WS29Bcfr7Bb9tZxA",
                    "publicKey": [
                        {
                            "id": "did:sov:1H6d1WS29Bcfr7Bb9tZxA#cSwsDbSW",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:1H6d1WS29Bcfr7Bb9tZxA",
                            "publicKeyBase58": "gMxdVMyF8RakaptxFsDzHAxfQ5iEKZkCEQcSwsDbSWf",
                        }
                    ],
                    "authentication": [],
                    "service": [
                        {
                            "id": "did:sov:1H6d1WS29Bcfr7Bb9tZxA;0f555a80-c950-4b2c-b8ec-632c72ffd780",
                            "type": "IndyAgent",
                            "priority": 0,
                            "recipientKeys": [
                                "gMxdVMyF8RakaptxFsDzHAxfQ5iEKZkCEQcSwsDbSWf"
                            ],
                            "serviceEndpoint": "http://172.17.0.1:9031",
                        }
                    ],
                },
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:1H6d1WS29Bcfr7Bb9tZxA",
                    "verificationMethod": [
                        {
                            "id": "did:sov:1H6d1WS29Bcfr7Bb9tZxA#cSwsDbSW",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:1H6d1WS29Bcfr7Bb9tZxA",
                            "publicKeyBase58": "gMxdVMyF8RakaptxFsDzHAxfQ5iEKZkCEQcSwsDbSWf",
                        }
                    ],
                    "authentication": [],
                    "service": [
                        {
                            "id": "did:sov:1H6d1WS29Bcfr7Bb9tZxA#didcomm-0",
                            "type": "did-communication",
                            "priority": 0,
                            "recipientKeys": ["did:sov:1H6d1WS29Bcfr7Bb9tZxA#cSwsDbSW"],
                            "serviceEndpoint": "http://172.17.0.1:9031",
                        }
                    ],
                },
            ),
            (  # Stored doc with routing keys
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP",
                    "publicKey": [
                        {
                            "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP#1",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:PkWfCgY4SSAYeSoaWx3RFP",
                            "publicKeyBase58": "DQBMbzLmAK5iyiuPnqNvfkx3CQ2iTJ2sKz98utvdET4K",
                        },
                        {
                            "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP#QPUs2fFT",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:PkWfCgY4SSAYeSoaWx3RFP",
                            "publicKeyBase58": "cK7fwfjpakMuv8QKVv2y6qouZddVw4TxZNQPUs2fFTd",
                        },
                    ],
                    "authentication": [
                        {
                            "type": "Ed25519SignatureAuthentication2018",
                            "publicKey": "did:sov:PkWfCgY4SSAYeSoaWx3RFP#1",
                        }
                    ],
                    "service": [
                        {
                            "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP;PkWfCgY4SSAYeSoaWx3RFP#IndyAgentService",
                            "type": "IndyAgent",
                            "priority": 0,
                            "recipientKeys": [
                                "DQBMbzLmAK5iyiuPnqNvfkx3CQ2iTJ2sKz98utvdET4K"
                            ],
                            "routingKeys": [
                                "cK7fwfjpakMuv8QKVv2y6qouZddVw4TxZNQPUs2fFTd"
                            ],
                            "serviceEndpoint": "https://aries-mediator-agent.vonx.io",
                        }
                    ],
                },
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP",
                    "verificationMethod": [
                        {
                            "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP#1",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:PkWfCgY4SSAYeSoaWx3RFP",
                            "publicKeyBase58": "DQBMbzLmAK5iyiuPnqNvfkx3CQ2iTJ2sKz98utvdET4K",
                        }
                    ],
                    "authentication": ["did:sov:PkWfCgY4SSAYeSoaWx3RFP#1"],
                    "service": [
                        {
                            "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP#didcomm-0",
                            "type": "did-communication",
                            "priority": 0,
                            "recipientKeys": ["did:sov:PkWfCgY4SSAYeSoaWx3RFP#1"],
                            "routingKeys": [
                                "did:key:z6Mkf4aAGBvBA8Eq2Qy714sspCPoj8uUupJpeaHLDkq3aUF1#z6Mkf4aAGBvBA8Eq2Qy714sspCPoj8uUupJpeaHLDkq3aUF1"
                            ],
                            "serviceEndpoint": "https://aries-mediator-agent.vonx.io",
                        }
                    ],
                },
            ),
            (  # Stored bifold doc
                {
                    "@context": "https://w3id.org/did/v1",
                    "publicKey": [
                        {
                            "id": "PkWfCgY4SSAYeSoaWx3RFP#1",
                            "controller": "PkWfCgY4SSAYeSoaWx3RFP",
                            "type": "Ed25519VerificationKey2018",
                            "publicKeyBase58": "DQBMbzLmAK5iyiuPnqNvfkx3CQ2iTJ2sKz98utvdET4K",
                        }
                    ],
                    "service": [
                        {
                            "id": "PkWfCgY4SSAYeSoaWx3RFP#IndyAgentService",
                            "serviceEndpoint": "https://aries-mediator-agent.vonx.io",
                            "type": "IndyAgent",
                            "priority": 0,
                            "recipientKeys": [
                                "DQBMbzLmAK5iyiuPnqNvfkx3CQ2iTJ2sKz98utvdET4K"
                            ],
                            "routingKeys": [
                                "cK7fwfjpakMuv8QKVv2y6qouZddVw4TxZNQPUs2fFTd"
                            ],
                        }
                    ],
                    "authentication": [
                        {
                            "publicKey": "PkWfCgY4SSAYeSoaWx3RFP#1",
                            "type": "Ed25519SignatureAuthentication2018",
                        }
                    ],
                    "id": "PkWfCgY4SSAYeSoaWx3RFP",
                },
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP",
                    "verificationMethod": [
                        {
                            "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP#1",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:PkWfCgY4SSAYeSoaWx3RFP",
                            "publicKeyBase58": "DQBMbzLmAK5iyiuPnqNvfkx3CQ2iTJ2sKz98utvdET4K",
                        }
                    ],
                    "authentication": ["did:sov:PkWfCgY4SSAYeSoaWx3RFP#1"],
                    "service": [
                        {
                            "id": "did:sov:PkWfCgY4SSAYeSoaWx3RFP#IndyAgentService",
                            "type": "did-communication",
                            "priority": 0,
                            "recipientKeys": ["did:sov:PkWfCgY4SSAYeSoaWx3RFP#1"],
                            "routingKeys": [
                                "did:key:z6Mkf4aAGBvBA8Eq2Qy714sspCPoj8uUupJpeaHLDkq3aUF1#z6Mkf4aAGBvBA8Eq2Qy714sspCPoj8uUupJpeaHLDkq3aUF1"
                            ],
                            "serviceEndpoint": "https://aries-mediator-agent.vonx.io",
                        }
                    ],
                },
            ),
            (  # Doc with multiple services
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                    "publicKey": [
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                            "publicKeyBase58": "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG",
                        }
                    ],
                    "authentication": [
                        {
                            "type": "Ed25519SignatureAuthentication2018",
                            "publicKey": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1",
                        }
                    ],
                    "service": [
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ;indy",
                            "type": "IndyAgent",
                            "priority": 0,
                            "recipientKeys": [
                                "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG"
                            ],
                            "routingKeys": [
                                "9NnKFUZoYcCqYC2PcaXH3cnaGsoRfyGgyEHbvbLJYh8j"
                            ],
                            "serviceEndpoint": "http://bob:3000",
                        },
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ;indy-ws",
                            "type": "IndyAgent",
                            "priority": 1,
                            "recipientKeys": [
                                "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG"
                            ],
                            "routingKeys": [
                                "9NnKFUZoYcCqYC2PcaXH3cnaGsoRfyGgyEHbvbLJYh8j"
                            ],
                            "serviceEndpoint": "ws://bob:3000",
                        },
                    ],
                },
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                    "verificationMethod": [
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                            "publicKeyBase58": "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG",
                        }
                    ],
                    "authentication": ["did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1"],
                    "service": [
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#didcomm-0",
                            "type": "did-communication",
                            "priority": 0,
                            "recipientKeys": ["did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1"],
                            "routingKeys": [
                                "did:key:z6Mknq3MqipEt9hJegs6J9V7tiLa6T5H5rX3fFCXksJKTuv7#z6Mknq3MqipEt9hJegs6J9V7tiLa6T5H5rX3fFCXksJKTuv7"
                            ],
                            "serviceEndpoint": "http://bob:3000",
                        },
                        {
                            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#didcomm-1",
                            "type": "did-communication",
                            "priority": 1,
                            "recipientKeys": ["did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1"],
                            "routingKeys": [
                                "did:key:z6Mknq3MqipEt9hJegs6J9V7tiLa6T5H5rX3fFCXksJKTuv7#z6Mknq3MqipEt9hJegs6J9V7tiLa6T5H5rX3fFCXksJKTuv7"
                            ],
                            "serviceEndpoint": "ws://bob:3000",
                        },
                    ],
                },
            ),
        ],
    )
    def test_corrections(self, input_doc: dict, expected: dict):
        actual = test_module.LegacyDocCorrections.apply(input_doc)
        assert actual == expected
        assert expected == test_module.LegacyDocCorrections.apply(expected)
        doc = pydid.deserialize_document(actual)
        assert doc.service
        assert isinstance(doc.service[0], pydid.DIDCommService)

    @pytest.mark.parametrize(
        ("input_doc", "expected"),
        [
            (
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "StwSYX1WFcJ7MBfYWxmuQ9",
                    "publicKey": [
                        {
                            "type": "Ed25519VerificationKey2018",
                            "id": "StwSYX1WFcJ7MBfYWxmuQ9#1",
                            "controller": "StwSYX1WFcJ7MBfYWxmuQ9",
                            "publicKeyBase58": "F7cEyTgzUbFwHsTwC2cK2Zy8bdraeoMY8921gyDmefwK",
                        }
                    ],
                    "authentication": [
                        {
                            "type": "Ed25519VerificationKey2018",
                            "publicKey": "StwSYX1WFcJ7MBfYWxmuQ9#1",
                        }
                    ],
                    "service": [
                        {
                            "type": "IndyAgent",
                            "id": "StwSYX1WFcJ7MBfYWxmuQ9#IndyAgentService",
                            "serviceEndpoint": "https://example.com/endpoint",
                            "recipientKeys": [
                                "F7cEyTgzUbFwHsTwC2cK2Zy8bdraeoMY8921gyDmefwK"
                            ],
                            "routingKeys": [
                                "did:key:z6Mko2LnynhGbkPQdZ3PQBUgCmrzdH9aJe7HTs4LKontx8Ge"
                            ],
                        }
                    ],
                },
                {
                    "@context": "https://w3id.org/did/v1",
                    "id": "did:sov:StwSYX1WFcJ7MBfYWxmuQ9",
                    "verificationMethod": [
                        {
                            "type": "Ed25519VerificationKey2018",
                            "id": "did:sov:StwSYX1WFcJ7MBfYWxmuQ9#1",
                            "controller": "did:sov:StwSYX1WFcJ7MBfYWxmuQ9",
                            "publicKeyBase58": "F7cEyTgzUbFwHsTwC2cK2Zy8bdraeoMY8921gyDmefwK",
                        }
                    ],
                    "authentication": ["did:sov:StwSYX1WFcJ7MBfYWxmuQ9#1"],
                    "service": [
                        {
                            "id": "did:sov:StwSYX1WFcJ7MBfYWxmuQ9#didcomm-0",
                            "type": "did-communication",
                            "serviceEndpoint": "https://example.com/endpoint",
                            "recipientKeys": ["did:sov:StwSYX1WFcJ7MBfYWxmuQ9#1"],
                            "routingKeys": [
                                "did:key:z6Mko2LnynhGbkPQdZ3PQBUgCmrzdH9aJe7HTs4LKontx8Ge#z6Mko2LnynhGbkPQdZ3PQBUgCmrzdH9aJe7HTs4LKontx8Ge"
                            ],
                        }
                    ],
                },
            )
        ],
    )
    def test_corrections_on_doc_as_received(self, input_doc: dict, expected: dict):
        parsed = DIDDoc.deserialize(input_doc)
        actual = test_module.LegacyDocCorrections.apply(parsed.serialize())
        assert actual == expected
        assert expected == test_module.LegacyDocCorrections.apply(expected)
        doc = pydid.deserialize_document(actual)
        assert doc.service
        assert isinstance(doc.service[0], pydid.DIDCommService)



import json
from unittest import TestCase, mock

import pytest

from ...utils.jwe import JweRecipient
from .. import crypto as test_module
from ..error import WalletError
from ..key_type import BLS12381G1, ED25519
from ..util import str_to_b64

SEED_B64 = "MDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDA="
SEED = "00000000000000000000000000000000"
MESSAGE = b"Hello World"


class TestCrypto(TestCase):
    def test_validate_seed(self):
        assert test_module.validate_seed(None) is None

        assert SEED.encode("ascii") == test_module.validate_seed(SEED_B64)
        assert SEED.encode("ascii") == test_module.validate_seed(SEED)

        with pytest.raises(test_module.WalletError) as excinfo:
            test_module.validate_seed({"bad": "seed"})
        assert "value is not a string or bytes" in str(excinfo.value)

        with pytest.raises(test_module.WalletError) as excinfo:
            test_module.validate_seed(f"{SEED}{SEED}")
        assert "value must be 32 bytes in length" in str(excinfo.value)

    def test_seeds_keys(self):
        assert len(test_module.seed_to_did(SEED)) in (22, 23)

        (public_key, secret_key) = test_module.create_keypair(test_module.ED25519)
        assert public_key
        assert secret_key

        assert test_module.sign_pk_from_sk(secret_key) in secret_key

    def test_decode_pack_message_x(self):
        with (
            mock.patch.object(
                test_module, "decode_pack_message_outer", mock.MagicMock()
            ) as mock_decode_outer,
            mock.patch.object(
                test_module, "extract_payload_key", mock.MagicMock()
            ) as mock_extract,
        ):
            mock_decode_outer.return_value = (b"wrapper", {"my": b"recip"}, True)

            with pytest.raises(ValueError) as excinfo:
                test_module.decode_pack_message(b"encrypted", lambda x: None)
            assert "No corresponding recipient key found" in str(excinfo.value)

            mock_extract.return_value = (b"payload_key", None)
            with pytest.raises(ValueError) as excinfo:
                test_module.decode_pack_message(b"encrypted", lambda x: b"recip_secret")
            assert "Sender public key not provided" in str(excinfo.value)

    def test_did_is_self_certified(self):
        did = "Av63wJYM7xYR4AiygYq4c3"
        verkey = "6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ"
        assert test_module.did_is_self_certified(did, verkey)
        verkey = "~PKAYz8Ev4yoQgr2LaMAWFx"
        assert test_module.did_is_self_certified(did, verkey)
        verkey = "ABUF7uxYTxZ6qYdZ4G9e1Gi"
        assert not test_module.did_is_self_certified(did, verkey)
        did = "6YnVN5Qdb6mqimTIQcQmSXrHXKdTEdRn5YHZReezUTvta"
        verkey = "6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ"
        assert not test_module.did_is_self_certified(did, verkey)

    def test_decode_pack_message_outer_x(self):
        with pytest.raises(ValueError) as excinfo:
            test_module.decode_pack_message_outer(json.dumps({"invalid": "content"}))
        assert "Invalid packed message" in str(excinfo.value)

        recips = str_to_b64(
            json.dumps(
                {
                    "enc": "xchacha20poly1305_ietf",
                    "typ": "JWM/1.0",
                    "alg": "Quadruple rot-13",
                    "recipients": "not a list",
                }
            )
        )
        with pytest.raises(ValueError) as excinfo:
            test_module.decode_pack_message_outer(
                json.dumps(
                    {
                        "protected": recips,
                        "iv": "MTIzNDU",
                        "tag": "MTIzNDU",
                        "ciphertext": "MTIzNDU",
                    }
                )
            )
        assert "Invalid packed message" in str(excinfo.value)

        recips = str_to_b64(
            json.dumps(
                {
                    "enc": "xchacha20poly1305_ietf",
                    "typ": "JWM/1.0",
                    "alg": "Quadruple rot-13",
                    "recipients": [{"encrypted_key": "MTIzNDU"}],
                }
            )
        )
        with pytest.raises(ValueError) as excinfo:
            test_module.decode_pack_message_outer(
                json.dumps(
                    {
                        "protected": recips,
                        "iv": "MTIzNDU",
                        "tag": "MTIzNDU",
                        "ciphertext": "MTIzNDU",
                    }
                )
            )
        assert "Unsupported pack algorithm" in str(excinfo.value)

    def test_extract_pack_recipients_x(self):
        with pytest.raises(ValueError) as excinfo:
            test_module.extract_pack_recipients([JweRecipient(encrypted_key=b"")])
        assert "Blank recipient key" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            test_module.extract_pack_recipients(
                [JweRecipient(encrypted_key=b"MTIzNDU", header={"kid": "4mZ5TYv4oN"})] * 2
            )
        assert "Duplicate recipient key" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            test_module.extract_pack_recipients(
                [
                    JweRecipient(
                        encrypted_key=b"MTIzNDU",
                        header={"kid": "4mZ5TYv4oN", "sender": "4mZ5TYv4oN"},
                    )
                ]
            )
        assert "Missing iv" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            test_module.extract_pack_recipients(
                [
                    JweRecipient(
                        encrypted_key=b"MTIzNDU",
                        header={"kid": "4mZ5TYv4oN", "iv": "MTIzNDU"},
                    )
                ]
            )
        assert "Unexpected iv" in str(excinfo.value)

    def test_sign_ed25519_x_multiple_messages(self):
        with self.assertRaises(WalletError) as context:
            test_module.sign_message([b"message1", b"message2"], b"secret", ED25519)
        assert "ed25519 can only sign a single message" in str(context.exception)

    def test_sign_x_unsupported_key_type(self):
        with self.assertRaises(WalletError) as context:
            test_module.sign_message([b"message1", b"message2"], b"secret", BLS12381G1)
        assert "Unsupported key type: bls12381g1" in str(context.exception)

    def test_verify_ed25519_x_multiple_messages(self):
        with self.assertRaises(WalletError) as context:
            test_module.verify_signed_message(
                [b"message1", b"message2"], b"signature", b"verkey", ED25519
            )
        assert "ed25519 can only verify a single message" in str(context.exception)

    def test_verify_x_unsupported_key_type(self):
        with self.assertRaises(WalletError) as context:
            test_module.verify_signed_message(
                [b"message1", b"message2"], b"signature", b"verkey", BLS12381G1
            )
        assert "Unsupported key type: bls12381g1" in str(context.exception)



import json
from unittest import IsolatedAsyncioTestCase
from unittest.mock import ANY

from ...connections.models.conn_record import ConnRecord
from ...messaging.decorators.attach_decorator import AttachDecorator
from ...messaging.decorators.service_decorator import ServiceDecorator
from ...messaging.request_context import RequestContext
from ...protocols.out_of_band.v1_0.messages.invitation import InvitationMessage
from ...protocols.out_of_band.v1_0.models.oob_record import OobRecord
from ...storage.error import StorageNotFoundError
from ...tests import mock
from ...transport.inbound.receipt import MessageReceipt
from ...transport.outbound.message import OutboundMessage
from ...utils.testing import create_test_profile
from ..oob_processor import OobMessageProcessor, OobMessageProcessorError


class TestOobProcessor(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.inbound_message_router = mock.MagicMock()
        self.oob_processor = OobMessageProcessor(
            inbound_message_router=self.inbound_message_router
        )

        self.oob_record = mock.MagicMock(
            connection_id="a-connection-id",
            attach_thread_id="the-thid",
            their_service=ServiceDecorator(
                recipient_keys=["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
                routing_keys=["6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ"],
                endpoint="http://their-service-endpoint.com",
            ),
            emit_event=mock.CoroutineMock(),
            delete_record=mock.CoroutineMock(),
            save=mock.CoroutineMock(),
        )
        self.context = RequestContext.test_context(self.profile)
        self.context.message = InvitationMessage()

    async def test_clean_finished_oob_record_no_multi_use_no_request_attach(self):
        test_message = InvitationMessage()
        test_message.assign_thread_id("the-thid", "the-pthid")

        mock_oob = mock.MagicMock(
            emit_event=mock.CoroutineMock(),
            delete_record=mock.CoroutineMock(),
            multi_use=False,
            invitation=mock.MagicMock(requests_attach=[]),
        )

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=mock_oob),
        ) as mock_retrieve_oob:
            await self.oob_processor.clean_finished_oob_record(self.profile, test_message)

            assert mock_oob.state == OobRecord.STATE_DONE
            mock_oob.emit_event.assert_called_once()
            mock_oob.delete_record.assert_called_once()

            mock_retrieve_oob.assert_called_once_with(
                ANY, {"invi_msg_id": "the-pthid"}, {"role": OobRecord.ROLE_SENDER}
            )

    async def test_clean_finished_oob_record_multi_use(self):
        test_message = InvitationMessage()
        test_message.assign_thread_id("the-thid", "the-pthid")

        mock_oob = mock.MagicMock(
            emit_event=mock.CoroutineMock(),
            delete_record=mock.CoroutineMock(),
            multi_use=True,
            invitation=mock.MagicMock(requests_attach=[]),
        )

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=mock_oob),
        ) as mock_retrieve_oob:
            await self.oob_processor.clean_finished_oob_record(self.profile, test_message)

            mock_oob.emit_event.assert_called_once()
            mock_oob.delete_record.assert_not_called()

            mock_retrieve_oob.assert_called_once_with(
                ANY, {"invi_msg_id": "the-pthid"}, {"role": OobRecord.ROLE_SENDER}
            )

    async def test_clean_finished_oob_record_x(self):
        test_message = InvitationMessage()
        test_message.assign_thread_id("the-thid", "the-pthid")

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(),
        ) as mock_retrieve_oob:
            mock_retrieve_oob.side_effect = (StorageNotFoundError(),)

            await self.oob_processor.clean_finished_oob_record(self.profile, test_message)

    async def test_find_oob_target_for_outbound_message(self):
        mock_oob = mock.MagicMock(
            emit_event=mock.CoroutineMock(),
            delete_record=mock.CoroutineMock(),
            multi_use=True,
            invitation=mock.MagicMock(requests_attach=[]),
            invi_msg_id="the-pthid",
            our_recipient_key="3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
            their_service=ServiceDecorator(
                recipient_keys=["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
                endpoint="http://their-service-endpoint.com",
                routing_keys=["6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ"],
            ),
            our_service=ServiceDecorator(
                recipient_keys=["3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"],
                endpoint="http://our-service-endpoint.com",
                routing_keys=[],
            ),
        )

        message = json.dumps({"~thread": {"thid": "the-thid"}})
        outbound = OutboundMessage(reply_thread_id="the-thid", payload=message)

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=mock_oob),
        ) as mock_retrieve_oob:
            target = await self.oob_processor.find_oob_target_for_outbound_message(
                self.profile, outbound
            )

            assert target
            assert target.endpoint == "http://their-service-endpoint.com"
            assert target.recipient_keys == [
                "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"
            ]
            assert target.routing_keys == ["6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ"]
            assert target.sender_key == "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"

            payload = json.loads(outbound.payload)

            assert payload == {
                "~thread": {"thid": "the-thid", "pthid": "the-pthid"},
                "~service": {
                    "recipientKeys": ["3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"],
                    "serviceEndpoint": "http://our-service-endpoint.com",
                    "routingKeys": [],
                },
            }

            mock_retrieve_oob.assert_called_once_with(
                ANY, {"attach_thread_id": "the-thid"}
            )

    async def test_find_oob_target_for_outbound_message_oob_not_found(self):
        message = json.dumps({})
        outbound = OutboundMessage(reply_thread_id="the-thid", payload=message)

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(side_effect=(StorageNotFoundError(),)),
        ) as mock_retrieve_oob:
            target = await self.oob_processor.find_oob_target_for_outbound_message(
                self.profile, outbound
            )

            assert not target
            mock_retrieve_oob.assert_called_once_with(
                ANY, {"attach_thread_id": "the-thid"}
            )

    async def test_find_oob_target_for_outbound_message_update_service_thread(self):
        mock_oob = mock.MagicMock(
            emit_event=mock.CoroutineMock(),
            delete_record=mock.CoroutineMock(),
            multi_use=True,
            invitation=mock.MagicMock(requests_attach=[]),
            invi_msg_id="the-pthid",
            our_recipient_key="3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
            their_service=ServiceDecorator(
                recipient_keys=["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
                endpoint="http://their-service-endpoint.com",
                routing_keys=["6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ"],
            ),
            our_service=ServiceDecorator(
                recipient_keys=["3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"],
                endpoint="http://our-service-endpoint.com",
                routing_keys=[],
            ),
        )

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=mock_oob),
        ):
            message = json.dumps({})
            outbound = OutboundMessage(reply_thread_id="the-thid", payload=message)
            await self.oob_processor.find_oob_target_for_outbound_message(
                self.profile, outbound
            )
            payload = json.loads(outbound.payload)

            assert payload == {
                "~thread": {"pthid": "the-pthid"},
                "~service": {
                    "recipientKeys": ["3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"],
                    "serviceEndpoint": "http://our-service-endpoint.com",
                    "routingKeys": [],
                },
            }

            message = json.dumps(
                {
                    "~service": {"already": "present"},
                }
            )
            outbound = OutboundMessage(reply_thread_id="the-thid", payload=message)
            await self.oob_processor.find_oob_target_for_outbound_message(
                self.profile, outbound
            )
            payload = json.loads(outbound.payload)

            assert payload == {
                "~thread": {"pthid": "the-pthid"},
                "~service": {"already": "present"},
            }

    async def test_find_oob_record_for_inbound_message_parent_thread_id(self):
        # With pthid
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

        # With pthid, throws error
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(side_effect=(StorageNotFoundError(),)),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

        # Without pthid
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt()

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_not_called()

    async def test_find_oob_record_for_inbound_message_connectionless_retrieve_oob(
        self,
    ):
        # With thread_id and recipient_verkey
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid",
                recipient_verkey="our-recipient-key",
                sender_verkey=self.oob_record.their_service.recipient_keys[0],
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(
                ANY,
                {
                    "attach_thread_id": "the-thid",
                    "our_recipient_key": "our-recipient-key",
                },
            )

        # With thread_id and recipient_verkey, throws error
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(side_effect=(StorageNotFoundError(),)),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", recipient_verkey="our-recipient-key"
            )

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(
                ANY,
                {
                    "attach_thread_id": "the-thid",
                    "our_recipient_key": "our-recipient-key",
                },
            )

        # With connection
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=None),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", recipient_verkey="our-recipient-key"
            )
            self.context.connection_record = mock.MagicMock()

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_not_called()

        # Without thread_id and recipient_verkey
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=None),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt()

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_not_called()

    async def test_find_oob_record_for_inbound_message_sender_connection_id_no_match(
        self,
    ):
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.oob_record.role = OobRecord.ROLE_SENDER
            self.oob_record.state = OobRecord.STATE_AWAIT_RESPONSE
            self.context.connection_record = mock.MagicMock(
                connection_id="a-connection-id"
            )
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

        # Connection id is different
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.oob_record.role = OobRecord.ROLE_SENDER
            self.oob_record.state = OobRecord.STATE_ACCEPTED
            self.context.connection_record = mock.MagicMock(
                connection_id="another-connection-id"
            )
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

        # Connection id is not the same, state is not await response
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.oob_record.role = OobRecord.ROLE_SENDER
            self.oob_record.state = OobRecord.STATE_ACCEPTED
            self.context.connection_record = mock.MagicMock(
                connection_id="another-connection-id"
            )
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

        # Connection id is not the same, state is AWAIT_RESPONSE. oob has connection_id
        with (
            mock.patch.object(
                OobRecord,
                "retrieve_by_tag_filter",
                mock.CoroutineMock(return_value=self.oob_record),
            ) as mock_retrieve,
            mock.patch.object(
                ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(
                    return_value=mock.MagicMock(delete_record=mock.CoroutineMock())
                ),
            ) as mock_retrieve_conn,
        ):
            self.oob_record.role = OobRecord.ROLE_SENDER
            self.oob_record.state = OobRecord.STATE_AWAIT_RESPONSE
            self.context.connection_record = mock.MagicMock(
                connection_id="another-connection-id"
            )
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})
            mock_retrieve_conn.assert_called_once_with(ANY, "a-connection-id")
            mock_retrieve_conn.return_value.delete_record.assert_called_once()

            assert self.oob_record.connection_id == "another-connection-id"

    async def test_find_oob_record_for_inbound_message_attach_thread_id_set(self):
        # No attach thread_id
        self.oob_record.attach_thread_id = None

        self.oob_record.invitation.requests_attach = [
            AttachDecorator.data_json({"@id": "the-thid"})
        ]

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

            assert self.oob_record.attach_thread_id == "the-thid"

    async def test_find_oob_record_for_inbound_message_attach_thread_id_not_in_list(
        self,
    ):
        # No attach thread_id
        self.oob_record.attach_thread_id = None

        self.oob_record.invitation.requests_attach = [
            AttachDecorator.data_json({"@id": "another-thid"})
        ]

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

    async def test_find_oob_record_for_inbound_message_not_attach_thread_id_matching(
        self,
    ):
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

    async def test_find_oob_record_for_inbound_message_not_attach_thread_id_not_matching(
        self,
    ):
        self.oob_record.attach_thread_id = "another-thid"

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid", parent_thread_id="the-pthid"
            )

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

    async def test_find_oob_record_for_inbound_message_recipient_verkey_not_in_their_service(
        self,
    ):
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid",
                parent_thread_id="the-pthid",
                recipient_verkey="recipient-verkey",
                sender_verkey="a-sender-verkey",
            )

            assert not await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

    async def test_find_oob_record_for_inbound_message_their_service_matching_with_message_receipt(
        self,
    ):
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid",
                parent_thread_id="the-pthid",
                recipient_verkey="recipient-verkey",
                sender_verkey="9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC",
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

    async def test_find_oob_record_for_inbound_message_their_service_set_on_oob_record(
        self,
    ):
        self.context._message._service = ServiceDecorator(
            endpoint="http://example.com/endpoint",
            recipient_keys=["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
            routing_keys=["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
        )

        self.oob_record.their_service = None

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid",
                parent_thread_id="the-pthid",
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

            assert self.oob_record.their_service == {
                "serviceEndpoint": "http://example.com/endpoint",
                "recipientKeys": ["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
                "routingKeys": ["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
            }

    async def test_find_oob_record_for_inbound_message_session_emit_delete(
        self,
    ):
        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid",
                parent_thread_id="the-pthid",
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

            assert self.oob_record.state == OobRecord.STATE_DONE
            self.oob_record.emit_event.assert_called_once()
            self.oob_record.delete_record.assert_called_once()
            self.oob_record.save.assert_not_called()

    async def test_find_oob_record_for_inbound_message_session_connectionless_save(
        self,
    ):
        self.oob_record.connection_id = None

        with mock.patch.object(
            OobRecord,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(return_value=self.oob_record),
        ) as mock_retrieve:
            self.context.message_receipt = MessageReceipt(
                thread_id="the-thid",
                parent_thread_id="the-pthid",
            )

            assert await self.oob_processor.find_oob_record_for_inbound_message(
                self.context
            )
            mock_retrieve.assert_called_once()
            mock_retrieve.assert_called_once_with(ANY, {"invi_msg_id": "the-pthid"})

            self.oob_record.emit_event.assert_not_called()
            self.oob_record.delete_record.assert_not_called()
            self.oob_record.save.assert_called_once()

    async def test_handle_message_connection(self):
        oob_record = mock.MagicMock(
            connection_id="the-conn-id",
            save=mock.CoroutineMock(),
            attach_thread_id=None,
            their_service=None,
        )

        await self.oob_processor.handle_message(
            self.profile,
            [
                {
                    "@type": "issue-credential/1.0/offer-credential",
                    "@id": "4a580490-a9d8-44f5-a3f6-14e0b8a219b0",
                }
            ],
            oob_record,
            their_service=ServiceDecorator(
                endpoint="http://their-service-endpoint.com",
                recipient_keys=["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
                routing_keys=["6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ"],
            ),
        )

        assert oob_record.attach_thread_id is None
        assert oob_record.their_service is None

        oob_record.save.assert_not_called()

        self.inbound_message_router.assert_called_once_with(self.profile, ANY, False)

    async def test_handle_message_connectionless(self):
        oob_record = mock.MagicMock(save=mock.CoroutineMock(), connection_id=None)

        await self.oob_processor.handle_message(
            self.profile,
            [
                {
                    "@type": "issue-credential/1.0/offer-credential",
                    "@id": "4a580490-a9d8-44f5-a3f6-14e0b8a219b0",
                }
            ],
            oob_record,
            their_service=ServiceDecorator(
                endpoint="http://their-service-endpoint.com",
                recipient_keys=["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
                routing_keys=["6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ"],
            ),
        )

        assert oob_record.attach_thread_id == "4a580490-a9d8-44f5-a3f6-14e0b8a219b0"
        assert oob_record.their_service.serialize() == {
            "serviceEndpoint": "http://their-service-endpoint.com",
            "recipientKeys": ["9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"],
            "routingKeys": ["6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ"],
        }

        oob_record.save.assert_called_once()

        self.inbound_message_router.assert_called_once_with(self.profile, ANY, False)

    async def test_handle_message_unsupported_message_type(self):
        with self.assertRaises(OobMessageProcessorError) as err:
            await self.oob_processor.handle_message(
                self.profile, [{"@type": "unsupported"}], mock.MagicMock()
            )
        assert (
            "None of the oob attached messages supported. Supported message types are issue-credential/1.0/offer-credential, issue-credential/2.0/offer-credential, present-proof/1.0/request-presentation, present-proof/2.0/request-presentation"
            in err.exception.message
        )

    async def test_get_thread_id(self):
        message_w_thread = {
            "@id": "the-message-id",
            "~thread": {"thid": "the-thread-id"},
        }
        message_wo_thread = {"@id": "the-message-id"}

        assert self.oob_processor.get_thread_id(message_w_thread) == "the-thread-id"
        assert self.oob_processor.get_thread_id(message_wo_thread) == "the-message-id"



"""Utility functions for the admin server."""

import re
from hmac import compare_digest


def const_compare(string1, string2):
    """Compare two strings in constant time."""
    if string1 is None or string2 is None:
        return False
    return compare_digest(string1.encode(), string2.encode())


def strip_did_prefix(did: str) -> str | None:
    """Strip the DID prefix from a DID."""
    if did:
        return re.sub(r"^did:\w+:", "", did)



"""Verify Leaf Inclusion."""

from .hasher import TreeHasher


class MerkleVerifier:
    """Utility class for verifying leaf inclusion."""

    def __init__(self, hasher=TreeHasher()):
        """Initialize MerkleVerifier."""
        self.hasher = hasher

    async def calculate_root_hash(
        self,
        leaf,
        leaf_index,
        audit_path,
        tree_size,
    ):
        """Calculate root hash, used to verify Merkel AuditPath.

        Reference: section 2.1.1 of RFC6962.

        Args:
            leaf: Leaf data.
            leaf_index: Index of the leaf in the tree.
            audit_path: A list of SHA-256 hashes representing the Merkle audit
            path.
            tree_size: tree size

        """
        leaf_hash = self.hasher.hash_leaf(leaf)
        if leaf_index >= tree_size or leaf_index < 0:
            return False
        fn, sn = leaf_index, tree_size - 1
        r = leaf_hash
        for p in audit_path:
            if self.lsb(fn) or (fn == sn):
                r = self.hasher.hash_children(p, r)
                while not ((fn == 0) or self.lsb(fn)):
                    fn >>= 1
                    sn >>= 1
            else:
                r = self.hasher.hash_children(r, p)
            fn >>= 1
            sn >>= 1
        return r

    def lsb(self, x):
        """Return Least Significant Bits."""
        return x & 1



TEST_LD_DOCUMENT_REVEAL = {
    "@context": ["http://schema.org/", "https://w3id.org/security/bbs/v1"],
    "@type": "Person",
    "@explicit": True,
    "firstName": "Jane",
    "lastName": "Does",
}



from unittest import TestCase

from ..key_type import (
    BLS12381G1,
    BLS12381G1G2,
    BLS12381G2,
    ED25519,
    P256,
    X25519,
    KeyTypes,
)

ED25519_PREFIX_BYTES = b"\xed\x01"
BLS12381G1_PREFIX_BYTES = b"\xea\x01"
BLS12381G1G2_PREFIX_BYTES = b"\xee\x01"
BLS12381G2_PREFIX_BYTES = b"\xeb\x01"
X25519_PREFIX_BYTES = b"\xec\x01"
P256_PREFIX_BYTES = b"\x80\x24"

ED25519_KEY_NAME = "ed25519"
X25519_KEY_NAME = "x25519"
P256_KEY_NAME = "p256"
BLS12381G1_KEY_NAME = "bls12381g1"
BLS12381G2_KEY_NAME = "bls12381g2"
BLS12381G1G2_KEY_NAME = "bls12381g1g2"

ED25519_MULTICODEC_NAME = "ed25519-pub"
X25519_MULTICODEC_NAME = "x25519-pub"
P256_MULTICODEC_NAME = "p256-pub"
BLS12381G1_MULTICODEC_NAME = "bls12_381-g1-pub"
BLS12381G2_MULTICODEC_NAME = "bls12_381-g2-pub"
BLS12381G1G2_MULTICODEC_NAME = "bls12_381-g1g2-pub"

ED25519_JWS_ALG = "EdDSA"
X25519_JWS_ALG = None
P256_JWS_ALG = "ES256"
BLS12381G1_JWS_ALG = None
BLS12381G2_JWS_ALG = None
BLS12381G1G2_JWS_ALG = None


class TestKeyType(TestCase):
    def test_from_multicodec_name(self):
        key_types = KeyTypes()
        assert key_types.from_multicodec_name(ED25519_MULTICODEC_NAME) == ED25519
        assert key_types.from_multicodec_name(X25519_MULTICODEC_NAME) == X25519
        assert key_types.from_multicodec_name(P256_MULTICODEC_NAME) == P256
        assert key_types.from_multicodec_name(BLS12381G1_MULTICODEC_NAME) == BLS12381G1
        assert key_types.from_multicodec_name(BLS12381G2_MULTICODEC_NAME) == BLS12381G2
        assert (
            key_types.from_multicodec_name(BLS12381G1G2_MULTICODEC_NAME) == BLS12381G1G2
        )
        assert key_types.from_multicodec_name("non-existing") is None

    def test_from_key_type(self):
        key_types = KeyTypes()
        assert key_types.from_key_type(ED25519_KEY_NAME) == ED25519
        assert key_types.from_key_type(X25519_KEY_NAME) == X25519
        assert key_types.from_key_type(P256_KEY_NAME) == P256
        assert key_types.from_key_type(BLS12381G1_KEY_NAME) == BLS12381G1
        assert key_types.from_key_type(BLS12381G2_KEY_NAME) == BLS12381G2
        assert key_types.from_key_type(BLS12381G1G2_KEY_NAME) == BLS12381G1G2
        assert key_types.from_key_type("non-existing") is None

    def test_from_multicodec_prefix(self):
        key_types = KeyTypes()
        assert key_types.from_multicodec_prefix(ED25519_PREFIX_BYTES) == ED25519
        assert key_types.from_multicodec_prefix(X25519_PREFIX_BYTES) == X25519
        assert key_types.from_multicodec_prefix(P256_PREFIX_BYTES) == P256
        assert key_types.from_multicodec_prefix(BLS12381G1_PREFIX_BYTES) == BLS12381G1
        assert key_types.from_multicodec_prefix(BLS12381G2_PREFIX_BYTES) == BLS12381G2
        assert key_types.from_multicodec_prefix(BLS12381G1G2_PREFIX_BYTES) == BLS12381G1G2
        assert key_types.from_multicodec_prefix(b"\xef\x01") is None

    def test_from_prefixed_bytes(self):
        key_types = KeyTypes()
        assert (
            key_types.from_prefixed_bytes(
                b"".join([ED25519_PREFIX_BYTES, b"random-bytes"])
            )
            == ED25519
        )
        assert (
            key_types.from_prefixed_bytes(
                b"".join([X25519_PREFIX_BYTES, b"random-bytes"])
            )
            == X25519
        )
        assert (
            key_types.from_prefixed_bytes(b"".join([P256_PREFIX_BYTES, b"random-bytes"]))
            == P256
        )
        assert (
            key_types.from_prefixed_bytes(
                b"".join([BLS12381G1_PREFIX_BYTES, b"random-bytes"])
            )
            == BLS12381G1
        )
        assert (
            key_types.from_prefixed_bytes(
                b"".join([BLS12381G2_PREFIX_BYTES, b"random-bytes"])
            )
            == BLS12381G2
        )
        assert (
            key_types.from_prefixed_bytes(
                b"".join([BLS12381G1G2_PREFIX_BYTES, b"random-bytes"])
            )
            == BLS12381G1G2
        )
        assert (
            key_types.from_prefixed_bytes(b"".join([b"\xef\x01", b"other-random-bytes"]))
            is None
        )

    def test_properties(self):
        key_type = ED25519

        assert key_type.key_type == ED25519_KEY_NAME
        assert key_type.multicodec_name == ED25519_MULTICODEC_NAME
        assert key_type.multicodec_prefix == ED25519_PREFIX_BYTES
        assert key_type.jws_algorithm == ED25519_JWS_ALG

        key_type = X25519

        assert key_type.key_type == X25519_KEY_NAME
        assert key_type.multicodec_name == X25519_MULTICODEC_NAME
        assert key_type.multicodec_prefix == X25519_PREFIX_BYTES
        assert key_type.jws_algorithm == X25519_JWS_ALG

        key_type = P256

        assert key_type.key_type == P256_KEY_NAME
        assert key_type.multicodec_name == P256_MULTICODEC_NAME
        assert key_type.multicodec_prefix == P256_PREFIX_BYTES
        assert key_type.jws_algorithm == P256_JWS_ALG

        key_type = BLS12381G1

        assert key_type.key_type == BLS12381G1_KEY_NAME
        assert key_type.multicodec_name == BLS12381G1_MULTICODEC_NAME
        assert key_type.multicodec_prefix == BLS12381G1_PREFIX_BYTES
        assert key_type.jws_algorithm == BLS12381G1_JWS_ALG

        key_type = BLS12381G2

        assert key_type.key_type == BLS12381G2_KEY_NAME
        assert key_type.multicodec_name == BLS12381G2_MULTICODEC_NAME
        assert key_type.multicodec_prefix == BLS12381G2_PREFIX_BYTES
        assert key_type.jws_algorithm == BLS12381G2_JWS_ALG

        key_type = BLS12381G1G2

        assert key_type.key_type == BLS12381G1G2_KEY_NAME
        assert key_type.multicodec_name == BLS12381G1G2_MULTICODEC_NAME
        assert key_type.multicodec_prefix == BLS12381G1G2_PREFIX_BYTES
        assert key_type.jws_algorithm == BLS12381G1G2_JWS_ALG



import pytest
import pytest_asyncio

from ...core.profile import Profile
from ...messaging.responder import BaseResponder, MockResponder
from ...protocols.coordinate_mediation.v1_0.models.mediation_record import MediationRecord
from ...protocols.coordinate_mediation.v1_0.route_manager import RouteManager
from ...protocols.routing.v1_0.manager import RoutingManager
from ...protocols.routing.v1_0.models.route_record import RouteRecord
from ...storage.error import StorageNotFoundError
from ...tests import mock
from ...utils.testing import create_test_profile
from ..base import BaseMultitenantManager
from ..route_manager import BaseWalletRouteManager, MultitenantRouteManager

TEST_RECORD_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_VERKEY = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
TEST_ROUTE_RECORD_VERKEY = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"
TEST_ROUTE_VERKEY = "did:key:z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya"
TEST_ROUTE_VERKEY_REF = "did:key:z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya#z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya"
TEST_ROUTE_VERKEY_REF2 = "did:key:z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhyz#z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhyz"


@pytest.fixture
def mock_responder():
    yield MockResponder()


@pytest_asyncio.fixture
async def root_profile(mock_responder: MockResponder):
    profile = await create_test_profile()
    profile.context.injector.bind_instance(BaseResponder, mock_responder)
    yield profile


@pytest_asyncio.fixture
async def sub_profile(mock_responder: MockResponder):
    profile = await create_test_profile()
    profile.context.injector.bind_instance(BaseResponder, mock_responder)
    yield profile


@pytest.fixture
def route_manager(root_profile: Profile, sub_profile: Profile):
    yield MultitenantRouteManager(root_profile)


@pytest.fixture
def base_route_manager():
    yield BaseWalletRouteManager()


@pytest.mark.asyncio
async def test_route_for_key_sub_mediator_no_base_mediator(
    route_manager: MultitenantRouteManager,
    mock_responder: MockResponder,
    sub_profile: Profile,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", connection_id="test-mediator-conn-id"
    )

    with (
        mock.patch.object(
            route_manager,
            "get_base_wallet_mediator",
            mock.CoroutineMock(return_value=None),
        ),
        mock.patch.object(
            RoutingManager, "create_route_record", mock.CoroutineMock()
        ) as mock_create_route_record,
    ):
        keylist_update = await route_manager._route_for_key(
            sub_profile,
            TEST_VERKEY,
            mediation_record,
            skip_if_exists=False,
            replace_key=None,
        )

    mock_create_route_record.assert_called_once_with(
        recipient_key=TEST_VERKEY, internal_wallet_id=sub_profile.name
    )
    assert keylist_update
    assert keylist_update.serialize()["updates"] == [
        {"action": "add", "recipient_key": TEST_VERKEY}
    ]
    assert mock_responder.messages
    assert (
        keylist_update,
        {"connection_id": "test-mediator-conn-id"},
    ) == mock_responder.messages[0]


@pytest.mark.asyncio
async def test_route_for_key_sub_mediator_and_base_mediator(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
    mock_responder: MockResponder,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", connection_id="test-mediator-conn-id"
    )
    base_mediation_record = MediationRecord(
        mediation_id="test-base-mediation-id",
        connection_id="test-base-mediator-conn-id",
    )

    with (
        mock.patch.object(
            route_manager,
            "get_base_wallet_mediator",
            mock.CoroutineMock(return_value=base_mediation_record),
        ),
        mock.patch.object(
            RoutingManager, "create_route_record", mock.CoroutineMock()
        ) as mock_create_route_record,
    ):
        keylist_update = await route_manager._route_for_key(
            sub_profile,
            TEST_VERKEY,
            mediation_record,
            skip_if_exists=False,
            replace_key=None,
        )

    mock_create_route_record.assert_called_once_with(
        recipient_key=TEST_VERKEY, internal_wallet_id=sub_profile.name
    )
    assert keylist_update
    assert keylist_update.serialize()["updates"] == [
        {"action": "add", "recipient_key": TEST_VERKEY}
    ]
    assert mock_responder.messages
    assert (
        keylist_update,
        {"connection_id": "test-base-mediator-conn-id"},
    ) == mock_responder.messages[0]


@pytest.mark.asyncio
async def test_route_for_key_base_mediator_no_sub_mediator(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
    mock_responder: MockResponder,
):
    base_mediation_record = MediationRecord(
        mediation_id="test-base-mediation-id",
        connection_id="test-base-mediator-conn-id",
    )

    with (
        mock.patch.object(
            route_manager,
            "get_base_wallet_mediator",
            mock.CoroutineMock(return_value=base_mediation_record),
        ),
        mock.patch.object(
            RoutingManager, "create_route_record", mock.CoroutineMock()
        ) as mock_create_route_record,
    ):
        keylist_update = await route_manager._route_for_key(
            sub_profile,
            TEST_VERKEY,
            None,
            skip_if_exists=False,
            replace_key=None,
        )

    mock_create_route_record.assert_called_once_with(
        recipient_key=TEST_VERKEY, internal_wallet_id=sub_profile.name
    )
    assert keylist_update
    assert keylist_update.serialize()["updates"] == [
        {"action": "add", "recipient_key": TEST_VERKEY}
    ]
    assert mock_responder.messages
    assert (
        keylist_update,
        {"connection_id": "test-base-mediator-conn-id"},
    ) == mock_responder.messages[0]


@pytest.mark.asyncio
async def test_route_for_key_skip_if_exists_and_exists(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
    mock_responder: MockResponder,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", connection_id="test-mediator-conn-id"
    )
    with mock.patch.object(
        RouteRecord, "retrieve_by_recipient_key", mock.CoroutineMock()
    ):
        keylist_update = await route_manager._route_for_key(
            sub_profile,
            TEST_VERKEY,
            mediation_record,
            skip_if_exists=True,
            replace_key=None,
        )
    assert keylist_update is None
    assert not mock_responder.messages


@pytest.mark.asyncio
async def test_route_for_key_skip_if_exists_and_absent(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
    mock_responder: MockResponder,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", connection_id="test-mediator-conn-id"
    )
    with mock.patch.object(
        RouteRecord,
        "retrieve_by_recipient_key",
        mock.CoroutineMock(side_effect=StorageNotFoundError),
    ):
        keylist_update = await route_manager._route_for_key(
            sub_profile,
            TEST_VERKEY,
            mediation_record,
            skip_if_exists=True,
            replace_key=None,
        )
    assert keylist_update
    assert keylist_update.serialize()["updates"] == [
        {"action": "add", "recipient_key": TEST_VERKEY}
    ]
    assert mock_responder.messages
    assert (
        keylist_update,
        {"connection_id": "test-mediator-conn-id"},
    ) == mock_responder.messages[0]


@pytest.mark.asyncio
async def test_route_for_key_replace_key(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
    mock_responder: MockResponder,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", connection_id="test-mediator-conn-id"
    )
    keylist_update = await route_manager._route_for_key(
        sub_profile,
        TEST_VERKEY,
        mediation_record,
        skip_if_exists=False,
        replace_key=TEST_ROUTE_VERKEY,
    )
    assert keylist_update
    assert keylist_update.serialize()["updates"] == [
        {"action": "add", "recipient_key": TEST_VERKEY},
        {"action": "remove", "recipient_key": TEST_ROUTE_VERKEY},
    ]
    assert mock_responder.messages
    assert (
        keylist_update,
        {"connection_id": "test-mediator-conn-id"},
    ) == mock_responder.messages[0]


@pytest.mark.asyncio
async def test_route_for_key_no_mediator(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
):
    assert (
        await route_manager._route_for_key(
            sub_profile,
            TEST_VERKEY,
            None,
            skip_if_exists=True,
            replace_key=TEST_ROUTE_VERKEY,
        )
        is None
    )


@pytest.mark.asyncio
async def test_routing_info_with_mediator(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id",
        connection_id="test-mediator-conn-id",
        routing_keys=[TEST_ROUTE_VERKEY_REF],
        endpoint="http://mediator.example.com",
    )
    keys, endpoint = await route_manager.routing_info(sub_profile, mediation_record)
    assert keys == mediation_record.routing_keys
    assert endpoint == mediation_record.endpoint


@pytest.mark.asyncio
async def test_routing_info_no_mediator(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
):
    keys, endpoint = await route_manager.routing_info(sub_profile, None)
    assert keys is None
    assert endpoint is None


@pytest.mark.asyncio
async def test_routing_info_with_base_mediator(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
):
    base_mediation_record = MediationRecord(
        mediation_id="test-base-mediation-id",
        connection_id="test-base-mediator-conn-id",
        routing_keys=[TEST_ROUTE_VERKEY_REF],
        endpoint="http://base.mediator.example.com",
    )

    with mock.patch.object(
        route_manager,
        "get_base_wallet_mediator",
        mock.CoroutineMock(return_value=base_mediation_record),
    ):
        keys, endpoint = await route_manager.routing_info(sub_profile, None)
    assert keys == base_mediation_record.routing_keys
    assert endpoint == base_mediation_record.endpoint


@pytest.mark.asyncio
async def test_routing_info_with_base_mediator_and_sub_mediator(
    sub_profile: Profile,
    route_manager: MultitenantRouteManager,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id",
        connection_id="test-mediator-conn-id",
        routing_keys=[TEST_ROUTE_VERKEY_REF2],
        endpoint="http://mediator.example.com",
    )
    base_mediation_record = MediationRecord(
        mediation_id="test-base-mediation-id",
        connection_id="test-base-mediator-conn-id",
        routing_keys=[TEST_ROUTE_VERKEY_REF],
        endpoint="http://base.mediator.example.com",
    )

    with mock.patch.object(
        route_manager,
        "get_base_wallet_mediator",
        mock.CoroutineMock(return_value=base_mediation_record),
    ):
        keys, endpoint = await route_manager.routing_info(sub_profile, mediation_record)
    assert keys == [*base_mediation_record.routing_keys, *mediation_record.routing_keys]
    assert endpoint == mediation_record.endpoint


@pytest.mark.asyncio
async def test_connection_from_recipient_key(
    sub_profile: Profile, base_route_manager: BaseWalletRouteManager
):
    manager = mock.MagicMock(BaseMultitenantManager, autospec=True)
    manager.get_profile_for_key = mock.CoroutineMock(return_value=sub_profile)
    sub_profile.context.injector.bind_instance(BaseMultitenantManager, manager)
    with mock.patch.object(
        RouteManager, "connection_from_recipient_key", mock.CoroutineMock()
    ) as mock_conn_for_recip:
        result = await base_route_manager.connection_from_recipient_key(
            sub_profile, TEST_VERKEY
        )
        assert result == mock_conn_for_recip.return_value



from unittest import IsolatedAsyncioTestCase, mock

import pytest

from .....did.did_key import DIDKey
from .....utils.testing import create_test_profile
from .....wallet.base import BaseWallet
from .....wallet.key_type import BLS12381G2
from ....tests.data import (
    TEST_LD_DOCUMENT_BAD_PARTIAL_PROOF_BBS,
    TEST_LD_DOCUMENT_PARTIAL_PROOF_BBS,
    TEST_LD_DOCUMENT_PROOF_BBS,
    TEST_LD_DOCUMENT_REVEAL,
    TEST_LD_DOCUMENT_REVEAL_ALL,
    TEST_LD_DOCUMENT_SIGNED_BBS,
    TEST_VC_DOCUMENT_NESTED_PARTIAL_PROOF_BBS,
    TEST_VC_DOCUMENT_NESTED_PROOF_BBS,
    TEST_VC_DOCUMENT_NESTED_REVEAL,
    TEST_VC_DOCUMENT_NESTED_SIGNED_BBS,
    TEST_VC_DOCUMENT_PARTIAL_PROOF_BBS,
    TEST_VC_DOCUMENT_REVEAL,
    TEST_VC_DOCUMENT_SIGNED_BBS,
)
from ....tests.document_loader import custom_document_loader
from ....vc_ld import derive_credential, verify_credential
from ...crypto.wallet_key_pair import WalletKeyPair
from ...error import LinkedDataProofException
from ...ld_proofs import derive, verify
from ...purposes.assertion_proof_purpose import AssertionProofPurpose
from ..bbs_bls_signature_proof_2020 import BbsBlsSignatureProof2020


@pytest.mark.ursa_bbs_signatures
class TestBbsBlsSignatureProof2020(IsolatedAsyncioTestCase):
    test_seed = "testseed000000000000000000000001"

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            self.key = await wallet.create_signing_key(
                key_type=BLS12381G2, seed=self.test_seed
            )
        self.verification_method = DIDKey.from_public_key_b58(
            self.key.verkey, BLS12381G2
        ).key_id

        self.key_pair = WalletKeyPair(profile=self.profile, key_type=BLS12381G2)

    async def test_derive_ld_proofs(self):
        derived = await derive(
            document=TEST_LD_DOCUMENT_SIGNED_BBS,
            reveal_document=TEST_LD_DOCUMENT_REVEAL,
            document_loader=custom_document_loader,
            suite=BbsBlsSignatureProof2020(key_pair=self.key_pair),
        )

        assert derived

    async def test_verify_derived_x_bad_proof(self):
        result = await verify(
            document=TEST_LD_DOCUMENT_BAD_PARTIAL_PROOF_BBS,
            suites=[BbsBlsSignatureProof2020(key_pair=self.key_pair)],
            document_loader=custom_document_loader,
            purpose=AssertionProofPurpose(),
        )

        assert not result.verified

    async def test_derive_x_unsigned_claims(self):
        doc = {**TEST_LD_DOCUMENT_SIGNED_BBS, "unsigned_claim": True}

        with self.assertRaises(Exception) as context:
            await derive(
                document=doc,
                reveal_document=TEST_LD_DOCUMENT_REVEAL_ALL,
                document_loader=custom_document_loader,
                suite=BbsBlsSignatureProof2020(key_pair=self.key_pair),
            )
        assert "The messages and signature do not match" in str(context.exception)

    async def test_derive_x_modified_claims(self):
        doc = {**TEST_LD_DOCUMENT_SIGNED_BBS, "email": "bad@example.com"}

        with self.assertRaises(Exception) as context:
            await derive(
                document=doc,
                reveal_document=TEST_LD_DOCUMENT_REVEAL_ALL,
                document_loader=custom_document_loader,
                suite=BbsBlsSignatureProof2020(key_pair=self.key_pair),
            )
        assert "The messages and signature do not match" in str(context.exception)

    async def test_derive_x_removed_claims(self):
        doc = TEST_LD_DOCUMENT_SIGNED_BBS.copy()
        doc.pop("email")

        with self.assertRaises(Exception) as context:
            await derive(
                document=doc,
                reveal_document=TEST_LD_DOCUMENT_REVEAL_ALL,
                document_loader=custom_document_loader,
                suite=BbsBlsSignatureProof2020(key_pair=self.key_pair),
            )
        assert "The messages and signature do not match" in str(context.exception)

    async def test_derive_ld_proofs_reveal_all(self):
        derived = await derive(
            document=TEST_LD_DOCUMENT_SIGNED_BBS,
            reveal_document=TEST_LD_DOCUMENT_REVEAL_ALL,
            document_loader=custom_document_loader,
            suite=BbsBlsSignatureProof2020(key_pair=self.key_pair),
        )

        assert derived

    async def test_derive_vc(self):
        derived = await derive_credential(
            credential=TEST_VC_DOCUMENT_SIGNED_BBS,
            reveal_document=TEST_VC_DOCUMENT_REVEAL,
            document_loader=custom_document_loader,
            suite=BbsBlsSignatureProof2020(key_pair=self.key_pair),
        )

        assert derived

    async def test_derive_vc_nested(self):
        derived = await derive_credential(
            credential=TEST_VC_DOCUMENT_NESTED_SIGNED_BBS,
            reveal_document=TEST_VC_DOCUMENT_NESTED_REVEAL,
            document_loader=custom_document_loader,
            suite=BbsBlsSignatureProof2020(key_pair=self.key_pair),
        )

        assert derived

    async def test_verify_ld_proof(self):
        result = await verify(
            document=TEST_LD_DOCUMENT_PROOF_BBS,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
            suites=[BbsBlsSignatureProof2020(key_pair=self.key_pair)],
        )

        assert result.verified

    async def test_verify_derived_partial_ld_proof(self):
        result = await verify(
            document=TEST_LD_DOCUMENT_PARTIAL_PROOF_BBS,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
            suites=[BbsBlsSignatureProof2020(key_pair=self.key_pair)],
        )

        assert result.verified

    async def test_verify_derived_nested_vc_full_reveal(self):
        result = await verify_credential(
            credential=TEST_VC_DOCUMENT_NESTED_PROOF_BBS,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
            suites=[BbsBlsSignatureProof2020(key_pair=self.key_pair)],
        )

        assert result.verified

    async def test_verify_derived_nested_vc_partial_reveal(self):
        result = await verify_credential(
            credential=TEST_VC_DOCUMENT_NESTED_PARTIAL_PROOF_BBS,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
            suites=[BbsBlsSignatureProof2020(key_pair=self.key_pair)],
        )

        assert result.verified

    async def test_verify_derived_vc_partial_reveal(self):
        result = await verify_credential(
            credential=TEST_VC_DOCUMENT_PARTIAL_PROOF_BBS,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
            suites=[BbsBlsSignatureProof2020(key_pair=self.key_pair)],
        )

        assert result.verified

    async def test_derive_proof_x_invalid_proof_type(self):
        suite = BbsBlsSignatureProof2020(
            key_pair=self.key_pair,
        )

        with self.assertRaises(LinkedDataProofException):
            await suite.derive_proof(
                reveal_document=mock.MagicMock(),
                document=mock.MagicMock(),
                proof={"type": "incorrect type"},
                document_loader=mock.MagicMock(),
            )



TEST_LD_DOCUMENT_PROOF_BBS = {
    "@context": ["http://schema.org/", "https://w3id.org/security/v3-unstable"],
    "id": "urn:bnid:_:c14n0",
    "type": "Person",
    "email": "jane.doe@example.com",
    "firstName": "Jane",
    "jobTitle": "Professor",
    "lastName": "Does",
    "telephone": "(425) 123-4567",
    "proof": {
        "type": "BbsBlsSignatureProof2020",
        "nonce": "Qvv+50burAq9dHFIE6drJWKVcQj4G+IdwQXT/S2mHkxwbd4ZI/d4lrWQ1uiJPNToHx4=",
        "proofValue": "AAoD/5BmimoSIcEAL2qzBgzHMkRF5GiNLInT30Klkp/DKgzYdIxad+DDoKSLtu3HqUJZ44bT77EubVQclHcJl3zsI826G77z53am0jZRBCmVyTRqbjIbmlBGR+DlSPs/e7Vo35Q8hKq0wBCQBLwQ6+ArSshYVPD3fvTd6lsOpvz0fJPkDYmCXhu2w9vdXZYgsobsyAAAAHSNUZ27LZKohwy3M8OoNBk59TfX8sb/ss8QagDeQ2IOQ6Tb6tyRVeF0kkDNuYno4JAAAAACXYgOfWJ8s+gDYgb/7qXnJ8amB7hhUJxWMkFGQd1D7yopAqyp9S9lwqwY5Zi8YwxfO9JjX9g5TYBorYVooI25molA0BcvIu0IzX23j9GJ9gpJuTHCUKtizD/tgepu+H6ow/mbsIeTlpUlAlasX7NIhgAAAAJpOZ/vLm9Omn7GqKBnoxCMK+dqi4Hr9NsGNOSk2qnW9Erf74dJA+e4CRMfP4ZCbYbbx7Ex+4cckdqUH4WRby4M",
        "verificationMethod": "did:example:489398593#test",
        "proofPurpose": "assertionMethod",
        "created": "2020-10-30T04:34:10Z",
    },
}
